# -*- coding: utf-8 -*-
"""
    femagtools.dakotaout
    ~~~~~~~~~~~~~~~~~~~~

    parse dakota out file (helper module)

"""
import re
import logging

_numPattern = re.compile(r'\s+([+-]?\d+(?:\.\d+)?(?:[eE][+-]\d+)?)')
logger = logging.getLogger(__name__)


def _readSections(f):
    """return list of dakota out sections

    sections are split by empty lines

    Args:
      param f (file) dakota file to be read

    Returns:
      list of sections
    """

    section = []
    for line in f:
        if line.strip():
            section.append(line.rstrip())
        else:
            yield section
            section = []
    yield section


def read_moat_analysis(l):
    nums = _numPattern.findall(l)
    if len(nums) > 3:
        return [float(n) for n in nums[1:]]


def read_sample_moment(s):
    moments = []
    for l in s[2:]:
        nums = _numPattern.findall(l)
        if len(nums) > 3:
            name = l.split()[0]
            nums = _numPattern.findall(l[len(name):])
            moments.append(dict(
                name=name,
                mean=float(nums[0]),
                stddev=float(nums[1]),
                skewness=float(nums[2]),
                kurtosis=float(nums[3])))
    return 'moments', moments


def read_confidence(s):
    conf = []
    keys = ['LowerCI_Mean', 'UpperCI_Mean',
            'LowerCI_StdDev', 'UpperCI_StdDev']
    for l in s[2:]:
        nums = _numPattern.findall(l)
        if len(nums) > 3:
            name = l.split()[0]
            nums = _numPattern.findall(l[len(name):])

            d = {k: float(n) for k, n in zip(
                keys, nums)}
            d['name'] = name
            conf.append(d)
    return 'conf95', conf


def read_correlation(s):
    d = dict()
    k = ''
    for title in [('Simple Correlation', 'corr'),
                  ('Partial Correlation', 'corrpartial'),
                  ('Simple Rank Correlation', 'rank'),
                  ('Partial Rank Correlation', 'rankpartial')]:
        if s[0].startswith(title[0]):
            k = title[1]
            break
    if k:
        for l in s[2:]:
            nums = _numPattern.findall(l)
            if nums:
                d[l.split()[0]] = [float(n) for n in nums]
    return k, d


def read_pdf(s):
    pdf = []
    d = dict()
    name = ''
    for l in s[1:]:
        nums = _numPattern.findall(l)
        if len(nums) > 1:
            d['lower'].append(float(nums[0]))
            d['upper'].append(float(nums[1]))
            d['density'].append(float(nums[2]))
        elif l.startswith('PDF'):
            if d:
                pdf.append(d)
            d = dict(name=l.split()[-1][:-1],
                     lower=[], upper=[], density=[])
    return 'pdf', pdf


def read_cdf(s):
    cdf = []
    d = dict()
    name = ''
    for l in s[1:]:
        nums = _numPattern.findall(l)
        if len(nums) > 1:
            d['level'].append(float(nums[0]))
            d['p'].append(float(nums[1]))
        elif l.startswith('Cumulative'):
            if d:
                cdf.append(d)
            d = dict(name=l.split()[-1][:-1],
                     level=[], p=[])
    return 'cdf', cdf


def read_optsol(s):
    optsol = dict(parameters=[], objectives=[])
    for l in s[1:]:
        vn = l.split()
        if len(vn) == 2:
            nums = _numPattern.findall(f' {vn[0]}')
            if len(nums) == 1:
                optsol['parameters'].append((vn[1], float(nums[0])))
        elif len(vn) == 1:
            nums = _numPattern.findall(f' {vn[0]}')
            if len(nums) == 1:
                optsol['objectives'].append(float(nums[0]))

    return 'optsol', optsol


def read_dakota_out(filename):
    sectitles = [
        ('Sample moment statistics', read_sample_moment),
        ('95% confidence intervals', read_confidence),
        ('Probability Density Function', read_pdf),
        ('Level mappings', read_cdf),
        ('Simple Correlation', read_correlation),
        ('Partial Correlation', read_correlation),
        ('Simple Rank Correlation', read_correlation),
        ('Partial Rank Correlation', read_confidence),
        ('<<<<< Best parameters', read_optsol)]

    recs = dict()
    optsol = []
    with open(filename) as fp:
        stat = False
        for l in fp:
            if l.startswith('<<<<< Function'):
                break
        if fp and l:
            samples = int(_numPattern.findall(l)[0])
            recs['samples'] = samples
            for s in _readSections(fp):
                for t in sectitles:
                    if s:
                        if s[0].startswith(t[0]):
                            logger.info(s[0][:-1])
                            if t[1]:
                                r = t[1](s)
                                if r[0]:
                                    if r[0] == 'optsol':
                                        optsol.append(r[1])
                                    else:
                                        recs[r[0]] = r[1]
    if optsol:
        import numpy as np
        parnames = set([p[0] for o in optsol for p in o['parameters']])
        x = np.array([[p[1] for p in o['parameters']] for o in optsol]).T
        f = np.array([[f for f in o['objectives']] for o in optsol]).T
        recs['opt'] = dict(parnames=list(parnames), x=x.tolist(), f=f.tolist())
    return recs


if __name__ == '__main__':
    import sys
    import logging.config
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')
    for a in sys.argv[1:]:
        print(read_dakota_out(a))



# -*- coding: utf-8 -*-
"""analyze windings

 Conventions

 Number of slots: Q
 Numper of pole pairs: p
 Number of phases: m
 Number of layers: l
 Number of wires per slot side: n
 Number of slots per pole and phase: q = Q/p/2/m
 Number of coils per phase: c = Q * l/2/m
 Number of parallel circuits (coil groups): g
 Number of turns per phase: w1 = Q * n * l/2/m/g
"""
import numpy as np
import femagtools.bch
from xml.etree import ElementTree as ET

coil_color = ['lime', 'gold', 'magenta',
              'blue', 'brown', 'blueviolet']


def num_basic_windings(Q, p, l):
    """return number of basic windings"""
    if l == 1:  # single layer
        return np.gcd(Q//2, p)
    return np.gcd(Q, p)


def q1q2yk(Q, p, m, l=1):
    """returns q1, q2, Yk, Qb"""
    t = num_basic_windings(Q, p, l)
    Qb = Q//t
    qqb = Qb if l == 2 else Qb//2
    pb = p//t
    if qqb//m % 2:  # odd
        q2 = (qqb + m)//(2*m) - 1
        q1 = q2 + 1
    else:
        q2 = (qqb)//(2*m)
        q1 = q2
    n = 1
    while (n*qqb + 1) % pb:
        n += 1
    Yk = (n*qqb + 1)//pb
    return q1, q2, Yk, Qb


class Winding(object):
    # TODO: raise ValueError "Unbalanced winding" if Q % (m * gcd(Q, p)) != 0

    def __init__(self, arg):
        """create winding either from bch winding section or winding dict()

        Arguments:
          arg: (object) femagtools.bch.Reader
            or dict Q: number of slots, m: number of phases, p: pole pairs, l: num layers, yd: coils span (slots)
       """
        if isinstance(arg, femagtools.bch.Reader):
            self.m = arg.machine['m']
            self.Q = arg.machine['Q']
            self.p = arg.machine['p']
            self.windings = arg.windings
        else:
            for k in arg.keys():
                setattr(self, k, arg[k])
        # balanced winding check
        if np.mod(self.Q, self.m*np.gcd(self.Q, self.p)):
            raise ValueError(
                f"Unbalanced winding: Q={self.Q}, p={self.p}, m={self.m}")
        self.q = self.Q/2/self.p/self.m  # number of coils per pole and phase

        if hasattr(self, 'windings'):
            # calculate coil width yd and num layers l
            taus = 360/self.Q
            ngen = 0
            if 'slots' in self.windings[1]:  # custom winding def
                for k in self.windings:
                    w = self.windings[k]
                    ngen = max(max(w['slots']), ngen)
                    w['dir'] = [1 if s > 0 else -1 for s in w['slots']]
                    w['PHI'] = [(2*abs(s)-1)*taus/2 for s in w['slots']]
                    w['R'] = [0 if l == 1 else 1 for l in w['layer']]

            self.yd = max(self.Q//self.p//2, 1)
            w = 1
            try:
                k = self.windings[w]['dir'].index(
                    -self.windings[w]['dir'][0])
                dphi = (self.windings[w]['PHI'][k] -
                        self.windings[w]['PHI'][0])
                self.yd = abs(round(dphi/taus))
            except ValueError:
                pass

            slots = [round((x-taus/2)/taus)
                     for x in self.windings[1]['PHI']]
            self.l = 2
            if len(slots) > ngen//self.m:
                self.l = 2
            return

        layers = 1
        if hasattr(self, 'l'):
            layers = self.l
        else:
            self.l = layers
        coilwidth = max(self.Q//self.p//2, 1)
        if hasattr(self, 'coilwidth'):   # obsolete, use yd instead
            coilwidth = self.coilwidth
        elif hasattr(self, 'yd'):
            coilwidth = self.yd

        self.yd = coilwidth

        q1, q2, Yk, Qb = q1q2yk(self.Q, self.p, self.m, self.l)
        k1 = [(q1 + q2)*i for i in range(self.m)]
        k2 = (q1*(self.m+1) + q2*(self.m-1))//2
        j = 2 if layers == 1 else 1
        pos = [[(j*Yk*(k + n)) % Qb
                for n in range(q1)] for k in k1]
        neg = [[j*Yk*(k + n + k2) % Qb
                for n in range(q2)] for k in k1]
        if layers > 1:
            slots = [sorted([(k, 1, 1)
                             for k in p] + [(k, -1, 1)
                                            for k in n])
                     for n, p in zip(neg, pos)]
            for i, p in enumerate(slots):
                slots[i] = sorted(slots[i] +
                                  [((k[0]+coilwidth) % Qb, -k[1], 0)
                                   for k in slots[i]], key=lambda s: s[0])
        else:
            if (coilwidth + 1) % 2:
                coilwidth += 1
            xneg = [sorted([s for s in n if s+1 % 2] +
                           [(s + coilwidth) % Qb for s in p if s+1 % 2])
                    for n, p in zip(neg, pos)]
            xpos = [sorted([s for s in p if s+1 % 2] +
                           [(s + coilwidth) % Qb for s in n if s+1 % 2])
                    for n, p in zip(neg, pos)]

            slots = [sorted([(k, 1, 1) for k in p] + [(k, -1, 1) for k in n])
                     for n, p in zip(xneg, xpos)]

        if self.m > 3:  # TODO: check this hack
            slots = slots[:self.m//2] + [[((k[0]+1) % self.Q, k[1], k[2])
                                          for k in s]
                                         for s in slots[:self.m//2]]
        taus = 360/self.Q
        self.windings = {
            i+1:  dict(dir=[k[1] for k in s],
                       N=[1]*len(s), R=[k[2] for k in s],
                       PHI=[taus/2+k[0]*taus for k in s])
            for i, s in enumerate(slots)}

    def kw_order(self, n):
        """return winding factor harmonics"""
        if n == 0:
            return self.p
        g = np.arange(-n, n, 1)
        t = num_basic_windings(self.Q, self.p, self.l)
        return self.p + g * self.m*t

    def kwp(self, n=0):
        """pitch factor"""
        if np.asarray(n).any() and not np.isscalar(n):
            nue = n
        else:
            nue = self.kw_order(n)
        return np.sin(nue*self.yd*np.pi/self.Q)

    def kwd(self, n=0):
        """zone (distribution) factor"""
        q1, q2, Yk, Qb = q1q2yk(self.Q, self.p, self.m, self.l)
        if np.asarray(n).any() and not np.isscalar(n):
            nue = n
        else:
            nue = self.kw_order(n)
        if q1 == q2:  # integral slot winding
            q = self.Q/2/self.m/self.p
            nuep = nue/self.p
            return np.sin(nuep*np.pi/2/self.m)/q/np.sin(nuep*np.pi/2/self.m/q)
        k = 2 if self.l == 1 else 1
        a = nue*k*np.pi/self.Q*Yk
        t = num_basic_windings(self.Q, self.p, self.l)
        return np.real((np.sin(a*q1) - np.exp(1j*np.pi/t*Yk*nue)*np.sin(a*q2))/((q1+q2)*np.sin(a)))

    def kw(self, n=0):
        """return winding factor"""
        return np.abs(self.kwp(n) * self.kwd(n))

    def harmleakcoeff(self, n=4000):
        """return harmonic leakage coefficient
        Arguments:
        n: (int) maximum number of harmonics"""
        nue = self.p * (1 + np.arange(1-n, n)*2*self.m)
        kw1 = self.kwd()*self.kwp()
        kwn = self.kwd(nue)*self.kwp(nue)
        return np.sum((self.p*kwn/nue/kw1)**2) - 1

    def coils_per_phase(self):
        """return number of coils per phase"""
        return self.Q * self.l/2/self.m

    def turns_per_phase(self, n, g):
        """return number of turns per phase
        Arguments:
        n: (int) number of wires per slot side
        g: (int) number of parallel coil groups
        """
        return n*self.coils_per_phase()//g

    def inductance(self, nwires, g, da1, lfe, ag, sw=0):
        """return main inductance / phase
        Arguments:
        nwires: number of wires in slot side
        g: (int) number of parallel groups
        da1: (float) bore diameter / m
        lfe: (float) length of lamination /m
        ag: (float) length of airgap / m
        sw: (float) slot opening width / m
        """
        mue0 = 4*np.pi*1e-7
        # carter factor
        tauq = np.pi*da1/self.Q
        h = sw/ag
        xic = 2/np.pi*(h*np.arctan(h/2) - np.log(1+(h/2)**2))
        kc = tauq/(tauq-xic*ag)  # Carter factor
        de = kc * ag

        N1 = self.turns_per_phase(nwires, g)
        return 3*mue0*da1*lfe*((self.kw()*N1)**2 /
                               (np.pi*self.p**2*de))

    def sequence(self):
        """returns sequence of winding keys"""
        return list(zip(*sorted([(k, self.windings[k]['PHI'][0])
                                 for k in self.windings.keys()],
                                key=lambda wdg: wdg[1])))[0]

    def slots(self, key):
        """returns slot indexes of winding key"""
        ngen = self.m*self.Q//np.gcd(self.Q, self.m*2*self.p)
        taus = 360/self.Q

        dim = int(self.l*ngen/self.m)
        slots = [(round((x-taus/2)/taus) + ngen*n) % self.Q + 1
                 for n in range(self.Q//ngen)
                 for x in self.windings[key]['PHI'][:dim]]
        return np.array(slots).reshape((np.gcd(self.Q, self.p), -1))

    def axis(self):
        """returns axis angle of winding 1 in mechanical system"""
        return self.mmf()['alfa0']

    def mmf(self, k=1, nmax=21):
        """returns the dimensionless magnetomotive force
        (ampere-turns/turns/ampere) and winding angle of phase k (rad)
        Arguments:
        k: (int) winding key
        nmax: (int) max order of harmonic (in electrical system)
        """
        slots = self.slots(k)[0]
        dirs = self.windings[k]['dir']
        # turns = self.windings[k]['N']
        r = len(slots)//len(dirs)
        curr = np.concatenate([np.array(dirs)*(1 - 2*(n % 2))
                               for n in range(r)])

        NY = 4096
        t = np.gcd(self.Q, self.p)
        y = np.zeros(NY*self.Q//t)
        for i in range(1, self.Q//t+1):
            if i in set(slots):
                y[NY*(i-1)+NY//2] = np.sum(curr[slots == i])
        yy = [np.sum(y[:i+1]) for i in range(0, len(y))]
        yy[:NY//2] = yy[-NY//2:]
        yy = np.tile(yy-np.mean(yy), t)
        yy /= np.max(yy)

        # calc spectrum
        N = len(yy)
        Y = np.fft.fft(yy)
        pb = self.p//t
        if self.q < 1:
            imax = pb
        else:
            imax = np.argmax(np.abs(Y[:N//2]))
        a = 2*np.abs(Y[imax])/N
        taus = 2*np.pi/self.Q
        freq = np.fft.fftfreq(N, d=taus/NY)
        T0 = np.abs(1/freq[imax])
        alfa0 = np.angle(Y[imax])
        # if alfa0 < 0: alfa0 += 2*np.pi
        pos_fft = np.linspace(0, 2*np.pi/t, 20*pb)
        D = (a*np.cos(2*np.pi*pos_fft/T0+alfa0))
        nue, mmf_nue = np.array(
            [(n, f) for n, f in zip(
                np.arange(0, nmax),
                2*np.abs(Y)/N) if f > 0]).T

        return dict(
            pos=[i*taus/NY for i in range(len(y))],
            mmf=yy[:NY*self.Q//t].tolist(),
            alfa0=-alfa0/self.p,
            nue=nue.tolist(),
            mmf_nue=mmf_nue.tolist(),
            pos_fft=pos_fft.tolist(),
            mmf_fft=D.tolist())

    def zoneplan(self):
        taus = 360/self.Q
        dphi = 1e-3
        slots = {k: [s-1 for s in self.slots(k)[0]]
                 for k in self.windings}
        layers = 1
        avgr = 0
        maxr, minr = max(self.windings[1]['R']), min(self.windings[1]['R'])
        if maxr-minr > 1e-6:
            layers = 2
            avgr = (maxr+minr)/2

            def is_upper(r, phi):
                return r > avgr
        elif len(slots[1]) > len(set(slots[1])):
            layers = 2

            def is_upper(r, phi):
                return phi < -dphi
        else:
            def is_upper(r, phi):
                return True

        upper = [[s+1 for s, x, r in zip(
            slots[key],
            self.windings[key]['PHI'],
            self.windings[key]['R'])
            if is_upper(r, s*taus - (x-taus/2))]
            for key in self.windings]
        udirs = [[d for s, d, x, r in zip(
            slots[key],
            self.windings[key]['dir'],
            self.windings[key]['PHI'],
            self.windings[key]['R'])
            if is_upper(r, s*taus - (x-taus/2))]
            for key in self.windings]
        lower = []
        ldirs = []
        if layers > 1:
            lower = [[s+1 for s, x, r in zip(
                slots[key],
                self.windings[key]['PHI'],
                self.windings[key]['R'])
                if not is_upper(r, s*taus - (x-taus/2))]
                for key in self.windings]
            ldirs = [[d for s, d, x, r in zip(
                slots[key],
                self.windings[key]['dir'],
                self.windings[key]['PHI'],
                self.windings[key]['R'])
                if not is_upper(r, s*taus - (x-taus/2))]
                for key in self.windings]

        z = ([[d*s for s, d in zip(u, ud)] for u, ud in zip(upper, udirs)],
             [[d*s for s, d in zip(l, ld)] for l, ld in zip(lower, ldirs)])
        # complete if not  basic winding:
        Qb = self.Q//num_basic_windings(self.Q, self.p, self.l)
        if max([abs(n) for m in z[0] for n in m]) < Qb:
            return [[k + [-n+Qb//2 if n < 0 else -(n+Qb//2) for n in k]
                     for k in m] for m in z]
        return z

    def diagram(self) -> ET.Element:
        """return winding diagram as svg element"""
        coil_len = 25
        coil_height = 4
        dslot = 8
        arrow_head_length = 2
        arrow_head_width = 2
        strokewidth = [f"{w}px" for w in [0.25, 0.5]]

        z = self.zoneplan()
        xoff = 0
        if z[-1]:
            xoff = 0.75
        yd = dslot*self.yd
        mh = 2*coil_height/yd
        slots = sorted([abs(n) for m in z[0] for n in m])
        smax = slots[-1]*dslot
        ET.register_namespace("", "http://www.w3.org/2000/svg")
        svg = ET.Element("svg", dict(
            version="1.1", xmlns="http://www.w3.org/2000/svg",
            viewBox=f"0, -30, {slots[-1] * dslot + 15}, 40"))
        g = ET.SubElement(svg, "g", {"id": "teeth", "fill": "lightblue"})
        for n in slots:
            e = ET.SubElement(g, "rect", {
                "x": f"{n * dslot + dslot/4}",
                "y": f"{-coil_len + 1}",
                "width": f"{dslot/2}",
                "height": f"{coil_len - 2}"})

        g = ET.SubElement(svg, "g",
                          {"id": "labels",
                           "text-anchor": "middle",
                           "dominant-baseline": "middle",
                           "style": "font-size: 0.15em; font-family: sans-serif;"})
        for n in slots:
            t = ET.SubElement(g, "text", {
                "x": f"{n*dslot}",
                "y": f"{-coil_len / 2}"}).text = str(n)

        g = ET.SubElement(svg, "g", {"id": "coils",
                                     "fill": "none",
                                     "stroke-linejoin": "round",
                                     "stroke-linecap": "round"})

        direction = ['right', 'left']
        nl = 2 if z[1] else 1
        for i, layer in enumerate(z):
            b = -xoff if i else xoff
            w = i if self.yd > 1 else 0
            for m, mslots in enumerate(layer):
                for k in mslots:
                    slotpos = abs(k) * dslot + b
                    pc = [f"L {slotpos} {-coil_len//2+2} M {slotpos} {-coil_len//2-1}",
                          f"L {slotpos} {-coil_len}"]

                    if nl == 2:
                        if k > 0:
                            d = 0 if i == 0 else 1
                        else:
                            d = 1 if i == 1 else 0
                    else:
                        d = 0 if k > 0 else 1

                    if direction[d] == 'right':
                        # first layer, positive dir or neg. dir and 2-layers:
                        #   from right bottom
                        if slotpos + yd > smax+b:
                            dx = dslot if yd > dslot else yd/4
                            ph = [f"M {slotpos+yd//2-xoff+dx} {coil_height-mh*dx}",
                                  f"L {slotpos+yd//2-xoff} {coil_height} L {slotpos} 0"]
                            pt = [f"L {slotpos+yd//2-xoff} {-coil_len-coil_height}",
                                  f"L {slotpos+yd//2-xoff+dx} {-coil_len-coil_height+mh*dx}"]
                        else:
                            ph = [
                                f"M {slotpos+yd//2-xoff} {coil_height} L {slotpos} 0"]
                            pt = [
                                f"L {slotpos+yd//2-xoff} {-coil_len-coil_height}"]
                    else:
                        # from left bottom
                        if slotpos - yd < 0:  # and slotpos - yd > -3*dslot:
                            dx = dslot if yd > dslot else yd/4
                            ph = [f"M {slotpos-yd//2+xoff-dx} {coil_height-mh*dx}",
                                  f"L {slotpos-yd//2+xoff} {coil_height} L {slotpos} 0"]
                            pt = [f"L {slotpos-yd//2+xoff} {-coil_len-coil_height}",
                                  f"L {slotpos-yd//2+xoff-dx} {-coil_len-coil_height+mh*dx}"]
                        else:
                            ph = [
                                f"M {slotpos-yd//2+xoff} {coil_height} L {slotpos} 0"]
                            pt = [
                                f"L {slotpos-yd//2+xoff} {-coil_len-coil_height}"]
                    e = ET.SubElement(g, "path", {
                        "d": ' '.join(ph + pc + pt),
                        "stroke-width": strokewidth[w],
                        "stroke": coil_color[m]})

        for i, layer in enumerate(z):
            for m, mslots in enumerate(layer):
                for k in mslots:
                    x = abs(k) * dslot
                    if i:
                        x -= xoff
                    else:
                        x += xoff
                    if k > 0:
                        y = coil_len * .88
                        points = [
                            (x, -y),
                            (x - arrow_head_width / 2, -y + arrow_head_length),
                            (x + arrow_head_width / 2, -y + arrow_head_length)]
                    else:
                        y = coil_len * .12
                        points = [
                            (x, -y),
                            (x - arrow_head_width / 2, -y - arrow_head_length),
                            (x + arrow_head_width / 2, -y - arrow_head_length)]
                    ET.SubElement(svg, "polygon", {
                        "points": " ".join([f"{x},{y}" for (x, y) in points]),
                        "fill": f"{coil_color[m]}",
                        "stroke": "none"})

        return ET.tostring(svg, encoding='unicode')

    def write(self, name, workdir='.'):
        """creates WID file"""
        import pathlib
        with open(pathlib.Path(workdir) / (name + '.WID'), 'w') as fp:
            if 'slots' in self.windings[1]:
                inp = sorted([(k if s > 0 else -k, abs(s), l, n)
                              for k in self.windings
                              for s, l, n in zip(
                    self.windings[k]['slots'],
                    self.windings[k]['layer'],
                    self.windings[k]['N'])],
                    key=lambda x: (x[1], x[2]))
                fp.write('\n'.join([
                    f'Windings input data: {name}',
                    ' Number of coil sides:',
                    f'          {len(inp)}',
                    'Type of machine: 1 = Rot, 21 = Lin-x, 22 = Lin-y',
                    '           1',
                    'Index  w-keys     N-turns         Layer', '']))

                for i, t in enumerate(inp):
                    fp.write(f'{i+1:5d} {t[0]:8d} {t[3]:10.3f} {t[2]:10d}\n')
            else:
                inp = sorted([(k if d > 0 else -k, r, phi, n)
                              for k in self.windings
                              for d, r, phi, n in zip(
                    self.windings[k]['dir'],
                    self.windings[k]['R'],
                    self.windings[k]['PHI'],
                    self.windings[k]['N'])],
                    key=lambda x: (x[1], x[2]))
                fp.write('\n'.join([
                    f'Windings input data: {name}',
                    ' Number of coil sides:',
                    f'          {len(inp)}',
                    'Type of machine: 1 = Rot, 21 = Lin-x, 22 = Lin-y',
                    '           1',
                    'Index  w-keys     N-turns         R[mm]     PHI[deg]', '']))
                for i, t in enumerate(inp):
                    fp.write(
                        f'{i+1:5d} {t[0]:8d} {t[3]:10.3f} {t[1]:8.3f} {t[2]:8.3f}\n')
            fp.write('\n'.join([
                'Number of windings saved :',
                f'         {len(self.windings)}',
                'W-Key Coil-Current [A]   W-Types: (=1 :wire&cur)'
            ] + [
                f'   {k}          0.00000000       0.00000000                            1'
                for k in self.windings
            ] + ['   0', '']))


if __name__ == "__main__":
    import sys
    import matplotlib.pyplot as plt
    from xml.etree import ElementTree as ET

    if sys.argv[1:]:
        bch = femagtools.bch.read(sys.argv[1])
        wdgs = Winding(bch)
    else:
        testdata = [
            dict(Q=90, p=12, m=3, l=2, coilwidth=1),
            dict(Q=54, p=6, m=3, l=2, coilwidth=5),
            dict(Q=168, p=7, m=3, l=2, coilwidth=10)]

        wdgs = Winding(testdata[1])

    c = wdgs.mmf()
    # print('alfa0={0:6.3f}'.format(wdgs.axis()/np.pi*180))

    plt.title('Q={0}, p={1}, alfa0={2:6.3f}'.format(
        wdgs.Q, wdgs.p, c['alfa0']/np.pi*180))
    plt.plot(np.array(c['pos'])/np.pi*180, c['mmf'])
    plt.plot(np.array(c['pos_fft'])/np.pi*180, c['mmf_fft'])

    phi = [c['alfa0']/np.pi*180, c['alfa0']/np.pi*180]
    y = [min(c['mmf_fft']), 1.1*max(c['mmf_fft'])]
    plt.plot(phi, y, '--')
    plt.annotate("", xy=(phi[0], y[0]),
                 xytext=(0, y[0]), arrowprops=dict(arrowstyle="->"))

    plt.grid()
    plt.show()

    svg = wdgs.diagram()
    tree = ET.ElementTree(ET.fromstring(svg))
    tree.write('wind.svg')
    print('SVG file "wind.svg" created')



import os
import sys
import json
from femagtools import amela
from pathlib import Path

def read_output():
    return (Path(__file__).parent/ "data/amela.out").read_text().split('\n')[0]

def read_json():
    return json.loads(
        (Path(__file__).parent/"data/pm_data/pm_data_se38.json").read_text())


def test_amela():
    al = amela.Amela(workdir='src/tests/data',
                     magnet_data=dict(name='pm_data'))
    loss = al()
    r = read_json()
    assert r['name'] == 'pm_data_se38'
    assert len(r['phi']) == 73
    assert read_output() == 'succeed'
    assert round(loss['pm_data_se38']['total_loss'], 3) == 0.881



import femagtools.femag
import femagtools.mcv

mcvData = [
    dict(curve=[dict(
        bi=[0.0, 0.45, 0.819, 1.196, 1.576, 1.81],
        hi=[0.0, 49.54, 95, 565.86, 15000, 20000])],
         name='m270-35a',
         desc=u"Demo Steel",
         ch=4.0,
         cw_freq=2.0,
         cw=1.68)]
mcv = femagtools.mcv.MagnetizingCurve(mcvData)

magnetmat = [
    dict(
        name='M45',
        remanenc=1.1,
        relperm=1.04,
        spmaweight=7.4,
        temcoefbr=-0.0015,
        temcoefhc=-0.0013,
        magncond=625000.0
    )]

machine = dict(
    name="PM 130 L4",
    outer_diam=0.13,
    bore_diam=0.07,
    inner_diam=0.,
    lfe=0.1,
    poles=4,
    stator=dict(
        mcvkey_yoke='m270-35a',
        num_slots=12,
        num_slots_gen=3,
        nodedist=1.5,
        rlength=1.0),
    magnet=dict(
        material='M45',
        magnetSector=dict(
            magn_num=1,
            magn_width_pct=0.8,
            magn_height=0.005,
            magn_shape=0.0,
            bridge_height=0.0,
            magn_type=1,
            condshaft_r=0.02,
            magn_ori=2,
            magn_rfe=0.0,
            bridge_width=0.0,
            magn_len=1.0)),
    windings=dict(
        num_phases=3,
        num_layers=1,
        num_wires=4,
        coil_span=3))

simulation = dict(
    calculationMode="cogg_calc",
    magn_temp=60.0,
    speed=50.0)


def test_run_script(monkeypatch, tmpdir):
    def mock_run(*args, **kwargs):
        return
    monkeypatch.setattr(femagtools.femag.Femag, "run", mock_run)
    femag = femagtools.femag.Femag(str(tmpdir),
                                   magnetizingCurves=mcv, magnets=magnetmat)

    fsl = femag.create_fsl(machine, simulation)
    assert len(fsl) == 94
    assert femag.model.magnet['temp_prop']['magntemp'] == 60.0

    r = femag(machine, simulation)

    assert isinstance(r, femagtools.bch.Reader)
    assert tmpdir.join("femag.fsl").exists()



"""
    femagtools.plot.forcedens
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Creating force density plots


"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm

def _create_3d_axis():
    """creates a subplot with 3d projection if one does not already exist"""
    from matplotlib.projections import get_projection_class
    from matplotlib import _pylab_helpers

    create_axis = True
    if _pylab_helpers.Gcf.get_active() is not None:
        if isinstance(plt.gca(), get_projection_class('3d')):
            create_axis = False
    if create_axis:
        plt.figure()
        plt.subplot(111, projection='3d')


def _plot_surface(ax, x, y, z, labels, azim=None):
    """helper function for surface plots"""
    # ax.tick_params(axis='both', which='major', pad=-3)
    assert np.size(x) > 1 and np.size(y) > 1 and np.size(z) > 1
    if azim is not None:
        ax.azim = azim
    X, Y = np.meshgrid(x, y)
    Z = np.ma.masked_invalid(z)
    ax.plot_surface(X, Y, Z,
                    rstride=1, cstride=1,
                    cmap=cm.viridis, alpha=0.85,
                    vmin=np.nanmin(z), vmax=np.nanmax(z),
                    linewidth=0, antialiased=True)
#                    edgecolor=(0, 0, 0, 0))
    # ax.set_xticks(xticks)
    # ax.set_yticks(yticks)
    # ax.set_zticks(zticks)
    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])
    ax.set_title(labels[2])


def forcedens(title, pos, fdens, ax=0):
    """plot force densities"""
    if ax == 0:
        ax = plt.gca()
    ax.set_title(title)
    ax.grid(True)

    ax.plot(pos, [1e-3*ft for ft in fdens[0]], label='F tang')
    ax.plot(pos, [1e-3*fn for fn in fdens[1]], label='F norm')
    ax.legend()
    ax.set_xlabel('Pos / deg')
    ax.set_ylabel('Force Density / kN/m²')


def forcedens_surface(fdens, ax=0):
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    xpos = [p for p in fdens.positions[0]['X']]
    ypos = [p['position'] for p in fdens.positions]
    z = 1e-3*np.array([p['FN']
                       for p in fdens.positions])
    _plot_surface(ax, xpos, ypos, z,
                  ('Rotor pos/°', 'Pos/°', 'F N / kN/m²'))


def forcedens_fft(title, fdens, nharm=40, ax=0):
    """plot force densities FFT
    Args:
      title: plot title
      fdens: force density object
      nharm: (int) num harmonics
    """
    if ax == 0:
        ax = plt.axes(projection="3d")

    F = 1e-3*fdens.fft(nharm)['fn_harm']['amplitude']
    fmin = 0.2
    num_bars = F.shape[0] + 1
    _xx, _yy = np.meshgrid(np.arange(1, num_bars),
                           np.arange(1, num_bars))
    z_size = F[F > fmin]
    x_pos, y_pos = _xx[F > fmin], _yy[F > fmin]
    z_pos = np.zeros_like(z_size)
    x_size = 2
    y_size = 2

    ax.bar3d(x_pos, y_pos, z_pos, x_size, y_size, z_size)
    #ax.view_init(azim=120)
    ax.set_xlim(0, num_bars+1)
    ax.set_ylim(0, num_bars+1)
    ax.set_title(title)
    ax.set_xlabel('M')
    ax.set_ylabel('N')
    ax.set_zlabel('kN/m²')



#!/usr/bin/env python
#
import os
import femagtools.tks


def test_read_tks():
    testPath = os.path.split(__file__)[0]
    if not testPath:
        testPath = '.'
    filename = "data/TKS-M400-65A.txt"
    tks = femagtools.tks.Reader('{0}/{1}'.format(testPath, filename))
    assert tks['name'] == 'TKS-M400-65A'
    assert len(tks['losses']['f']) == 4
    assert len(tks['losses']['B']) == 13
    assert len(tks['losses']['pfe']) == 4
    assert len(tks['losses']['pfe'][0]) == 13
    assert tks['losses']['f'] == [50.0, 100.0, 200.0, 400.0]
    assert round(tks['losses']['B'][-1], 1) == 1.8
    # 13.1279112, 17.92205587, 23.79869172, 30.84129721, 39.3434623, 48.97389844, 60.25629981, 72.88007909
    assert len([x for x in tks['losses']['pfe'][3] if x]) == 8



"""
    femagtools.moproblem
    ~~~~~~~~~~~~~~~~~~~~

    Creating and managing multi-objective problems



"""
from .moo.problem import Problem
import femagtools.magnet
import femagtools.bch
import femagtools.getset
import logging

logger = logging.getLogger(__name__)


class FemagMoProblem(Problem):
    """
    A multi-objective optimization problem with Femag.

    """

    def __init__(self, decision_vars, objective_vars):
        Problem.__init__(self, len(decision_vars), 0, len(objective_vars))
        if isinstance(decision_vars, dict):
            self.decision_vars = [{'name': n}
                                  for n in decision_vars['columns']]
            lbounds = [None]*len(self.decision_vars)
            ubounds = [None]*len(self.decision_vars)
        else:
            self.decision_vars = decision_vars
            lbounds, ubounds = zip(*[d['bounds'] if 'bounds' in d else (None, None)
                                     for d in decision_vars])
        self.set_bounds(lbounds, ubounds)

        logger.info("Decision Vars: %s", [
            d['name'] for d in self.decision_vars])
        logger.info("bounds lower: %s  upper: %s", lbounds, ubounds)
        self.objective_vars = objective_vars

    # prepare model
    def prepare(self, x, model):
        # list of dicts (model, feaModel, ..)
        if isinstance(model, list):
            for o in model:
                self.prepare(x, o)
            return

        # simple dict
        for d, v in zip(self.decision_vars, x):
            logger.debug("Prepare: %s = %s",  d['name'], v)
            attr = d['name'].split('.')
            if (attr[0] == 'magnetMat' and
                    isinstance(model, femagtools.magnet.Magnet)):
                try:
                    model.find(attr[1])[attr[2]] = v
                except TypeError:  # None
                    raise ValueError(f'Magnet material {attr[1]} not found')
            elif hasattr(model, attr[0]):
                model.set_value(attr, v)

    def setResult(self, result):
        self.result = result

    def objfun(self, x):
        if self.objective_vars:
            for o in self.objective_vars:
                logger.debug("%d=====> %s", len(self.objective_vars), str(o))
            return [f[0] * f[1] if f[1] is not None else None
                    for f in [(o.get('sign', 1),
                               self.result.get(o['name'].split('.')))
                              for o in self.objective_vars]]
        if (isinstance(self.result, femagtools.getset.GetterSetter) or
                isinstance(self.result, femagtools.bch.Reader)):
            return {k: v for k, v in self.result.items()}
        return self.result

    # Add some output to __repr__
    def __str__(self):
        return "\n\tMulti-Objective optimization problem"



import numpy as np

def fft(pos, y, pmod=0):
    """calculate fft spectrum of y and return samples,
    values, amplitude and phase of base harmonic

    Arguments:
      pos: (list of floats) sample positions
      y: (list of floats) y values
      pmod: number of poles in model (ignored if 0)
    """
    model_angle = pos[-1] - pos[0]
    ntiles = int(round(360/model_angle))

    if pmod:
        negative_periodic = pmod % 2
    else:
        negative_periodic = np.abs(y[0] - y[-1])/np.max(y) > 1

    if negative_periodic:
        yx = np.append(
            np.concatenate(
                [n*y[:-1]
                 for n in [m % 2 or -1
                           for m in range(1, ntiles+1)]]),
            y[0])
    else:
        yx = np.append(
            np.tile(y[:-1], ntiles),
            y[0])

    N = len(yx)
    # compute DFT from induction (eliminate DC offset)
    a0 = np.mean(yx)
    Y = np.fft.fft(yx-a0)

    # find the peak (amplitude of base harmonic)
    i = np.argmax(np.abs(Y[:N//2]))

    a = 2*np.abs(Y[i])/N
    freq = np.fft.fftfreq(N, d=pos[1]-pos[0])
    nmax = min(18*ntiles, N//2)
    T0 = 0
    if abs(freq[i]) > 0:
        T0 = np.abs(1/freq[i])
        npoles = 2*int(np.ceil(360/T0))
        nmax = min(9*npoles, N//2)

    alfa0 = np.angle(Y[i])

    return {'a': a, 'a0': a0, 'T0': T0, 'alfa0': alfa0,
            'nue': (2*np.abs(Y[:nmax])/N).tolist(),
            'yi': yx.tolist()}



# -*- coding: utf-8 -*-
""" Read ASM files

Slots Qr:

  Qr <= 1.25 Qs
  Qr != Qs +/- 2p
  Qr != Qs
  Qr != Qs +/- 1

"""
import pathlib
import logging
import logging.config
import numpy as np
import lmfit

logger = logging.getLogger('femagtools.asm')


def imcur(s, w1, u1, r1, ls1, lh, ls2, r2):
    """return currents i1r, i1i, i2r, i2i"""
    xs1 = w1*ls1
    xh = w1*lh
    x1 = xs1+xh
    xs2 = w1*ls2
    x2 = xs2+xh
    # solve voltage equations for steady state
    A = np.array((
        (r1, -x1, 0, -xh),
        (x1, r1, xh, 0),
        (0, -s*xh, r2, -s*x2),
        (s*xh, 0, s*x2, r2)))
    return np.linalg.solve(A, np.array((u1, 0, 0, 0)))


def torque(p, w1, u1, s, r1, ls1, lh, ls2, r2):
    """return torque"""
    if s == 0:
        return 0
    i2 = imcur(s, w1, u1, r1, ls1, lh, ls2, r2)[2:]
    return 3*p*r2/s/w1*(i2[0]**2 + i2[1]**2)


def fit_current(w1, u1, slip, r1, ls1, lh, i1, cosphi):
    def imcurr(s, ls2, r2):
        return np.array(
            [x + 1j*y
             for x, y in [imcur(sx, w1, u1, r1, ls1, lh, ls2, r2)[:2]
                          for sx in s]])
    model = lmfit.model.Model(imcurr)
    ls2_guess = 0.0
    r2_guess = np.mean(
        [u1/i1x*sx
         for i1x, sx in zip(i1, slip) if abs(i1x) > 1e-6])
    params = model.make_params(r2=r2_guess, ls2=ls2_guess)
    guess = lmfit.models.update_param_vals(params, model.prefix)
    i1c = np.array([x*pf - 1j*x*np.sqrt(1-pf**2)
                    for x, pf in zip(i1, cosphi)])
    r = model.fit(i1c, params=guess, s=slip, verbose=True)
    return r.params['ls2'].value, r.params['r2'].value


valmap = {
    'Stator windings voltage (RMS)[V]': 'u1',
    'Wdgs-connection: 0': 'wdgconn',
    'Nominal frequency': 'f1',
    'Stator phase winding resistamce': 'r1',
    'Stator phase winding resistance': 'r1',
    'Stator phase end-winding reactance': 'xs1',
    'Stator phase winding wires/slot side': 'num_wires',
    'Effect. air gap  length          [mm]': 'lfe',
    'Effect. rotor bar length (+end)  [mm]': 'rbarlen',
    'Number of Phases': 'num_phases',
    'Number of Pole pairs': 'p',
    'Number of Poles simulated': 'p_gen',
    'Number of parallel windings': 'num_par_wdgs',
    'Rotor Lamination                [kg]': 'lamweight',
    'Rotor Conductors                [kg]': 'conweight',
    'MC-File used in calculation': 'mcfile',
    'Losses[W/kg] in MC-File': 'felosscoeff',
    'Max. No. Iterations': 'maxiters',
    'Change of Perm. max %': 'permchg'}


def _read_sections(lines):
    """return list of ASM sections

    sections are either surrounded by lines starting with '[***'
    or by starting with any 'Input data' or 'Simulation Results'
    Args:
      param lines (list) lines of ASM file to read

    Returns:
      list of sections
    """

    section = []
    for line in lines:
        if ('[****' in line or
            'Input data' in line or
                'Simulation Results' in line):
            if section:
                # skip empty lines
                i = 0
                try:
                    while not section[i]:
                        i = i+1
                except IndexError:
                    i = i-1
                yield section[i:]
                section = []
        else:
            section.append(line.strip())
    yield section


def read_input_data(content):
    r = dict()
    for l in content:
        if '=' in l or ':' in l:
            d = '=' if '=' in l else ':'
            k, v = [s.strip() for s in l.split(d)[:2]]
            if k == 'Wdgs-connection: 0':
                c = int(float(l.split()[-1]))
                r[valmap[k]] = ['open', 'star', 'delta'][c]
            else:
                r[valmap[k]] = float(v)
        elif '\t' in l:
            k, v = [s.strip() for s in l.split('\t')[:2]]
            if valmap[k] == 'mcfile':
                r['mcfile'] = v
            else:
                r[valmap[k]] = float(v)

    for k in ('num_phases', 'p', 'p_gen',
              'num_par_wdgs', 'maxiters'):
        if k in r:
            r[k] = int(r[k])
    return r


def read_simulation_results(content):
    unit = 1e3
    resmap = {
        'Torque = P2/(s.omega) [Nm]': 'Tp2',
        'Rotor-Losses P2 [kW]': 'p2',
        'Rotor-Losses P2 [W]': 'p2',
        'Stator FE  Pfe1 [kW]': 'pfe1',
        'Stator FE  Pfe1 [W]': 'pfe1'}
    r = dict(s=[], T=[], u1=[], i1=[], p1=[], cosphi=[],
             f1=[], pfe1=[], p2=[], Tp2=[])
    for l in content:
        if l.startswith('S LIP'):
            if l.find('POWER[W]') > -1:
                unit = 1
                continue
        a = l.split()
        if len(a) == 7:
            for k, v in zip(r.keys(), a):
                r[k].append(float(v))
        elif a:
            a = l.split(':')[-1].split()
            if len(a) == 1:
                try:
                    k = resmap[l.split(':')[0]]
                    r[k].append(float(a[0]))
                except KeyError:
                    logger.warning('Key %s ignored', l.split(':')[0])
    if unit > 1:
        for k in ('p1', 'p2', 'pfe1'):
            r[k] = [x*unit for x in r[k]]
    r['s'] = [s/100 for s in r['s']]
    return r


def parident(w1, u1, i1, cosphi, s, r1, ls1, lh):
    """returns equivalent circuit parameters: r2, ls"""
    logger.info("w1 %s u1 %s i1 %s cosphi %s s %s r1 %s ls1 %s lh %s",
                w1, u1, i1, cosphi, s, r1, ls1, lh)
    ls2, r2 = fit_current(w1, u1, s,
                          r1, ls1, lh, i1, cosphi)

    xi = w1*(ls1+lh - lh**2/(ls2+lh))
    return dict(r2=r2, ls2=ls2, sk=r2/np.sqrt(xi**2 + r1**2))


def read(arg):
    """read asm file

        Args:
          filename or content (list of str) the text lines of the ASM file
    """
    r = {}
    if isinstance(arg, str):
        lines = pathlib.Path(arg).read_text().split('\n')
    elif isinstance(arg, pathlib.Path):
        lines = arg.read_text().split('\n')
    else:
        lines = arg
    for s in _read_sections(lines):
        if not s:
            continue
        title = s[0].split(':')[0].strip()

        if 'FEMAG Classic Version' in title:
            r['version'] = s[0].split(':')[-1].replace(' Version ', '').strip()
        elif 'Project File name' in title:
            r['project'] = s[1].strip()
        elif 'Number of Nodes' in title:
            pass
        elif 'File name' in title:
            r['filename'] = s[0].split(':')[-1].strip()
        elif 'Date' in title:
            d = s[0].split(':')[1].strip().split()
            dd, MM, yy = d[0].split('.')
            hh, mm = ''.join(d[1:-1]).split('.')
            r['date'] = '{}-{}-{}T{:02}:{:02}'.format(
                yy, MM, dd, int(hh), int(mm))
        elif 'Stator windings' in title:
            r.update(read_input_data(s))
        else:
            r.update(read_simulation_results(s))
    r['pcu'] = [3*x**2*r['r1'] for x in r['i1']]
    if r['pfe1']:
        r['pltotal'] = [sum(x) for x in zip(r['pcu'], r['pfe1'], r['p2'])]
    else:
        r['pltotal'] = [sum(x) for x in zip(r['pcu'], r['p2'])]

    w1 = 2*np.pi*r['f1'][0]
    r['ls1'] = r.pop('xs1')/w1
    u1 = r['u1'][0]
    i1 = r['i1']
    cosphi = r['cosphi']
    if r['wdgconn'] == 'star':
        u1 = u1/np.sqrt(3)
    if r['wdgconn'] == 'delta':
        i1 = [x/np.sqrt(3) for x in i1]

    r['lh'] = u1/i1[0]/w1 - r['ls1']
    if len(i1) > 2:
        r.update(parident(w1, u1, i1, cosphi,
                          r['s'], r['r1'], r['ls1'], r['lh']))
    return r


if __name__ == "__main__":
    import sys
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')
    r = read(sys.argv[1])
    print(r)



#!/usr/bin/env python
#
import unittest
import os
import femagtools.mcv


class McvReaderTest(unittest.TestCase):
    def test_read_mcv(self):
        testPath = os.path.split(__file__)[0]
        if not testPath:
            testPath = '.'
        filename = "data/TKS_NO_20.MCV"
        reader = femagtools.mcv.Reader()
        reader.readMcv('{0}/{1}'.format(testPath, filename))
        r = reader.get_results()
        self.assertEqual(r['desc'],
                         u'PowerCore\xae NO 20 ;ThyssenKrupp Steel Eur')
        self.assertEqual(len(r['curve'][0]['bi']), 24)
        self.assertEqual(r['curve'][0]['bi'][0], 0.0)
        self.assertAlmostEqual(r['curve'][0]['bi'][-1], 1.836, 3)
        
    def test_read_oriented_mcv(self):
        testPath = os.path.split(__file__)[0]
        if not testPath:
            testPath = '.'
        filename = "data/V800-50A_aniso.MCV"
        reader = femagtools.mcv.Reader()
        reader.readMcv('{0}/{1}'.format(testPath, filename))
        r = reader.get_results()
        self.assertEqual(r['curve'][0]['angle'], 0)
        self.assertEqual(r['curve'][1]['angle'], 90)
        self.assertEqual(len(r['curve'][0]['bi']), 17)
        self.assertEqual(r['curve'][0]['bi'][0], 0.0)
        self.assertAlmostEqual(r['curve'][0]['bi'][-1], 1.973, 3)
        self.assertAlmostEqual(r['curve'][1]['bi'][-1], 1.9763, 3)
        self.assertAlmostEqual(r['curve'][0]['hi'][-1], 16073, 0)
        self.assertAlmostEqual(r['curve'][1]['hi'][-1], 18718, 0)

    def test_read_PM_mcv(self):
        testPath = os.path.split(__file__)[0]
        if not testPath:
            testPath = '.'
        filename = "data/FERRIT_20gC.MCV"
        mcv = femagtools.mcv.read(
            '{0}/{1}'.format(testPath, filename))
        self.assertEqual(mcv.mc1_type, 2)
        self.assertAlmostEqual(min(mcv.curve[0]['hi']), -560e3)
        self.assertAlmostEqual(max(mcv.curve[0]['hi']), -50e3)
        self.assertAlmostEqual(min(mcv.curve[0]['bi']), -1.10371697)
        self.assertAlmostEqual(max(mcv.curve[0]['bi']), 0.4)

        
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
import pytest
import femagtools.machine.sizing


def test_im():
    P = 1.5e3
    fs = 12e3
    lda = 0.9
    p = 4
    f1 = 100
    n = f1/p
    udc = 550
    r = femagtools.machine.sizing.im(P, n, p, udc=udc,
                                     sigmas=fs, Ba=0.77,
                                     cos_phi=0.8, eta=0.8,
                                     lda=0.9)
    assert round(r['outer_diam'], 3) == 0.19
    assert r['stator']['num_slots'] == 36


def test_spm():
    P = 1.5e3
    fs = 12e3
    lda = 0.9
    p = 4
    f1 = 100
    n = f1/p
    udc = 550

    r = femagtools.machine.sizing.spm(P, n, p, udc=udc,
                                      Hc=700, sigmas=fs, brem=1.1, Ba=0.77,
                                      cos_phi=0.7, eta=0.8, demag=1.7,
                                      lda=0.9)
    assert round(r['outer_diam'], 3) == 0.19
    assert r['stator']['num_slots'] == 24


def test_eesm():
    P = 10e3
    speed = 4440/60
    p = 4
    udc = 600
    r = femagtools.machine.sizing.eesm(P, speed, p, udc=udc, Q1=36)
    assert r['rotor']['rot_hsm']



"""
    femagtools.plot.phasor
    ~~~~~~~~~~~~~~~~~~~~~~

    Creating phasor plots


"""
import numpy as np
import matplotlib.pyplot as plt


def __phasor_plot(ax, up, idq, uxdq):
    uref = max(up, uxdq[0])
    uxd = uxdq[0]/uref
    uxq = uxdq[1]/uref
    u1d, u1q = (uxd, 1+uxq)
    u1 = np.sqrt(u1d**2 + u1q**2)*uref
    i1 = np.linalg.norm(idq)
    i1d, i1q = (idq[0]/i1, idq[1]/i1)

    qhw = 6   # width arrow head
    qhl = 15  # length arrow head
    qlw = 2   # line width
    qts = 10  # textsize
    # Length of the Current adjust to Ud: Initally 0.9, Maier(Oswald) = 0.5
    curfac = max(0.9, 1.5*i1q/up)

    def label_line(ax, X, Y, U, V, label, color='k', size=8):
        """Add a label to a line, at the proper angle.

        Arguments
        ---------
        line : matplotlib.lines.Line2D object,
        label : str
        x : float
        x-position to place center of text (in data coordinated
        y : float
        y-position to place center of text (in data coordinates)
        color : str
        size : float
        """

        x1, x2 = X, X + U
        y1, y2 = Y, Y + V

        if y2 == 0:
            y2 = y1
        if x2 == 0:
            x2 = x1

        x = (x1 + x2) / 2
        y = (y1 + y2) / 2

        slope_degrees = np.rad2deg(np.angle(U + V * 1j))
        if slope_degrees < 0:
            slope_degrees += 180
        if 90 < slope_degrees <= 270:
            slope_degrees += 180

        x_offset = np.sin(np.deg2rad(slope_degrees))
        y_offset = np.cos(np.deg2rad(slope_degrees))
        bbox_props = dict(boxstyle="Round4, pad=0.1", fc="white", lw=0)
        text = ax.annotate(label, xy=(x, y), xytext=(x_offset * 10, y_offset * 8),
                           textcoords='offset points',
                           size=size, color=color,
                           horizontalalignment='center',
                           verticalalignment='center',
                           fontfamily='monospace', fontweight='bold', bbox=bbox_props)

        text.set_rotation(slope_degrees)
        return text

    if ax == 0:
        ax = plt.gca()
    ax.axes.xaxis.set_ticklabels([])
    ax.axes.yaxis.set_ticklabels([])
    # ax.set_aspect('equal')

    ax.set_title(
        r'$U_1$={0} V, $I_1$={1} A, $U_p$={2} V'.format(
            round(u1, 1), round(i1, 1), round(up, 1)), fontsize=14)

    up /= uref
    ax.quiver(0, 0, 0, up, angles='xy', scale_units='xy', scale=1, units='dots',
              headwidth=qhw/2, headlength=qhl/2, headaxislength=qhl/2, width=qlw*2, color='k')
    label_line(ax, 0, 0, 0, up, '$U_p$', 'k', qts)

    ax.quiver(0, 0, u1d, u1q, angles='xy', scale_units='xy', scale=1, units='dots',
              headwidth=qhw, headlength=qhl, headaxislength=qhl, width=qlw, color='r')
    label_line(ax, 0, 0, u1d, u1q, '$U_1$', 'r', qts)

    ax.quiver(0, 1, uxd, 0, angles='xy', scale_units='xy', scale=1, units='dots',
              headwidth=qhw, headlength=qhl, headaxislength=qhl, width=qlw, color='g')
    label_line(ax, 0, 1, uxd, 0, '$U_d$', 'g', qts)

    ax.quiver(uxd, 1, 0, uxq, angles='xy', scale_units='xy', scale=1, units='dots',
              headwidth=qhw, headlength=qhl, headaxislength=qhl, width=qlw, color='g')
    label_line(ax, uxd, 1, 0, uxq, '$U_q$', 'g', qts)

    ax.quiver(0, 0, curfac*i1d, curfac*i1q, angles='xy', scale_units='xy', scale=1,
              units='dots', headwidth=qhw, headlength=qhl, headaxislength=qhl, width=qlw, color='b')
    label_line(ax, 0, 0, curfac*i1d, curfac*i1q, '$I_1$', 'b', qts)

    xmin, xmax = (min(0, uxd, i1d), max(0, i1d, uxd))
    ymin, ymax = (min(0, i1q, 1-uxq), max(1, i1q, 1+uxq))

    ax.set_xlim([xmin-0.1, xmax+0.1])
    ax.set_ylim([ymin-0.1, ymax+0.1])
    ax.grid(True)


def i1beta_phasor(up, i1, beta, r1, xd, xq, ax=0):
    """creates a phasor plot
    up: internal voltage
    i1: current
    beta: angle i1 vs up [deg]
    r1: resistance
    xd: reactance in direct axis
    xq: reactance in quadrature axis"""

    i1d, i1q = (i1*np.sin(beta/180*np.pi), i1*np.cos(beta/180*np.pi))
    uxdq = ((r1*i1d - xq*i1q), (r1*i1q + xd*i1d))
    __phasor_plot(ax, up, (i1d, i1q), uxdq)


def iqd_phasor(up, iqd, uqd, ax=0):
    """creates a phasor plot
    up: internal voltage
    iqd: current
    uqd: terminal voltage"""

    uxdq = (uqd[1]/np.sqrt(2), (uqd[0]/np.sqrt(2)-up))
    __phasor_plot(ax, up, (iqd[1]/np.sqrt(2), iqd[0]/np.sqrt(2)), uxdq)


def phasor(bch, ax=0):
    """create phasor plot from bch"""
    f1 = bch.machine['p']*bch.dqPar['speed']
    w1 = 2*np.pi*f1
    xd = w1*bch.dqPar['ld'][-1]
    xq = w1*bch.dqPar['lq'][-1]
    r1 = bch.machine['r1']
    i1beta_phasor(bch.dqPar['up'][-1],
                  bch.dqPar['i1'][-1], bch.dqPar['beta'][-1],
                  r1, xd, xq, ax)



"""general design of an AC electrical machine

   Types: spm, ipm, eesm, im

"""
import numpy as np
import logging

logger = logging.getLogger("femagools.machineo.sizing")


PM_DEFAULTS = dict(
    airgap=1.5e-3,  # airgap width m
    eta=0.92,  # efficiency
    cos_phi=0.95,  # power factor
    m=3,  # number of phases
    ui_u=0.62,  # U ind / U
    lda=1.33,  # length/taup ratio 0.5 .. sqrt(p)
    J=3.8e6,  # current density 3 .. 6 A/mm²
    sigmas=17e3,  # shear force 10 .. 45 kN/m2
    Ba=0.7,  # flux density in airgap 0.6 .. 1.5 T
    Bth=1.5,  # flux density in teeth 1.5 .. 2 T
    By=1.2,  # flux density in yoke 1.2 .. 1.5 T
    kq=0.42,  # stator winding fill factor 0.35 .. 0.75
    mag_width=0.8,  # rel magnet width 0.6 .. 1
    Hc=700,  # max. coercitive field strength, 500 .. 900 kA/m
    brem=1.15,  # remanence 0.3 .. 1.3 T
    demag=1.5,  # safety factor for demagnetisation
    external_rotor=False,
    coil_span=0,
    hfe=1e-3  # iron height between magnet and airgap (IPM)
)
"""default sizing parameters for PM"""

SM_DEFAULTS = dict(
    airgap=1.5e-3,  # airgap width m
    eta=0.92,  # efficiency
    cos_phi=0.95,  # power factor
    m=3,  # number of phases
    ui_u=0.62,  # U ind / U
    lda=1.33,  # length/taup ratio 0.5 .. sqrt(p)
    J=3.8e6,  # current density 3 .. 6 A/mm²
    sigmas=17e3,  # shear force 10 .. 45 kN/m2
    Ba=0.7,  # flux density in airgap 0.6 .. 1.5 T
    Bth=1.5,  # flux density in stator teeth 1.5 .. 2 T
    By=1.2,  # flux density in stator yoke 1.2 .. 1.5 T
    Bthr=1.7,  # flux density in rotor teeth 1.5 .. 2 T
    Byr=1.4,  # flux density in rotor yoke 1.2 .. 1.5 T
    kq=0.42,  # stator winding fill factor 0.35 .. 0.75
    kqr=0.6,  # rotor winding fill factor 0.45 .. 0.7
    external_rotor=False,
    coil_span=0
)
"""default sizing parameters for SM"""

IM_DEFAULTS = dict(
    airgap=0.75e-3,  # airgap width m
    eta=0.87,  # efficiency
    cos_phi=0.8,  # power factor
    m=3,  # number of phases
    ui_u=0.62,  # U ind / U
    lda=1.33,  # length/taup ratio 0.5 .. sqrt(p)
    J=3e6,  # current density 3 .. 6 A/mm²
    sigmas=14e3,  # shear force 10 .. 45 kN/m2
    Ba=0.7,  # flux density in airgap 0.6 .. 1.5 T
    Bth=1.5,  # flux density in teeth 1.5 .. 2 T
    By=1.2,  # flux density in yoke 1.2 .. 1.5 T
    kq=0.42,  # winding fill factor 0.35 .. 0.75
    kqr=.9,   # bar fill factor
    external_rotor=False,
    coil_span=0
)
"""default sizing parameters for IM"""


class ValidationError(Exception):
    pass


def iron_losses(dy, lfe, hy, wt, ht, q1, B, f):
    dens = 7.8e3  # mass density of iron kg/m³
    ch = 3
    ce = 2
    Bo = 1.5
    fo = 50
    Wfe = ch*(B/Bo)**1.8*f/fo + ce*(B/Bo)**2*(f/fo)**2
    vol = lfe*(np.pi*(dy**2/4 - (dy/2-hy)**2) + q1*(ht*wt))
    return Wfe * dens*vol


def wdg_resistance(w1, l, d):
    S = 56e6  # conductivity of copper 1/Ohm m
    a = np.pi*d**2/4
    return 2*w1*l/S/a


def current_losses(w1, l, d, i1):
    # w1: num turns/phase
    # l: wire length/slot
    # d: wire diameter
    return wdg_resistance(w1, l, d) * i1**2


def check_symmetry_conditions(Q1, p, layers, m):
    # if Q1 % 2*p == 0:
    #    raise ValidationError("Number of slots Q and poles 2*p are not prime to each other")

    K = 2 if layers == 1 else 1
    urwick = np.gcd(Q1//K, p)
    if Q1 % m != 0:
        raise ValidationError(
            "Number of slots Q and phases are not consistent")

    K = 2 if m % 2 == 0 else 1
    if Q1 % K*m*urwick != 0:
        raise ValidationError(
            "Number of slots Q and phases are not consistent")


# number of rotor slots of IM:
# recommendations (Alger, Kuhlmann, Veinott):
# p, Q1 --> Qr
Q2R = {
    1: {36: [26, 28, 44],
        48: [38, 40, 56],
        54: [54],
        60: [52, 68, 78]},
    2: {36: [26, 28, 44],
        48: [34, 38, 40, 56],
        60: [34, 44, 46, 76],
        72: [58]},
    3: {36: [46, 48],
        48: [58, 60, 64, 68],
        54: [42, 66],
        72: [54, 58, 84, 88]},
    4: {36: [48, 52],
        48: [58, 64],
        54: [70],
        72: [58, 88]}}
# Richter:
Q2RR = {
    1: {24: [28, 16, 22],
        36: [24, 28, 48, 16],
        48: [40, 52],
        60: [48]},
    2: {36: [24, 40, 42, 60, 30, 44],
        48: [60, 84, 56, 44],
        60: [72, 48, 84, 44]},
    3: {36: [42, 48, 54, 30],
        54: [72, 88, 48],
        72: [96, 90, 84, 54]},
    4: {36: [48],
        48: [72, 60],
        72: [96, 84]}}

Q2IEC = {
    1: {18: [22],
        24: [18, 22],
        36: [],
        48: [],
        54: [],
        60: [],
        72: []}
}


def _im_stator_slots(p):
    return [k for k in Q2R[p].keys()]


def _rotor_slots(Q1, p):
    """returns a list of valid rotor slot numbers"""
    # Slots Qr of Induction Machines:
    #
    #  Qr <= 1.25 Qs
    #  Qr != Qs +/- 2p
    #  Qr != Qs
    #  Qr != Qs +/- 1
    # Slot area Acr: 0.7 - 0.9 Acs
    #
    # Examples of 4 poles SCIM Qs, Qr:
    # Audi e-Tron 48 58
    # Tesla S, X  60 74
    # Renault Twizy 36 48
    # others: 24/22, 36/24
    try:
        return Q2R[p][Q1]
    except KeyError:
        pass
    d = 0.25
    sync = Q1+p, Q1-p, Q1+2*p, Q1-2*p, Q1+5*p, Q1-5*p
    noise = Q1+1, Q1-1, Q1+2, Q1 + p+1, Q1-p-1, Q1+p+2, Q1-p-2
    return sorted(set(list(range(int((1-d)*Q1), int((1+d)*Q1)))) -
                  set((Q1, ) + sync + noise))


def _stator_slots(par, slots):
    """find a suitable number of slots
    (condition: 3 < slot height / width  < 5)"""
    q = []
    p = par['p']
    m = 3
    if slots:
        slotset = slots
    else:
        slotset = [n for n in range(12, 110, 2*m)
                   if n % (m*np.gcd(n, p)) == 0]
    for Q1 in slotset:
        par['Q1'] = Q1
        r = get_stator_dimensions(par)
        hb = r['hns']/r['bns']
        if hb > 5:
            break
        if hb > 3:
            q.append((Q1, Q1//np.gcd(Q1, p), abs(4-hb), r['kw']))

    if q:
        qhbmin = np.argmin(q, axis=0)
        logger.debug("q %s", q)
        # check sim factor, height/width ratio, winding factor
        if qhbmin[1] == len(q)-1:
            return q[qhbmin[1]][0]
        elif q[qhbmin[1]][2] < q[qhbmin[1]+1][2]:
            return q[qhbmin[1]][0]
        return q[qhbmin[1]+1][0]
    return slotset[0]


def get_stator_dimensions(par, slots=[]):
    # nominal (rated) operating parameters
    pnom = par['pnom']
    speednom = par['speed']
    tnom = pnom/(2*np.pi*speednom)
    p = par['p']  # pole pairs
    lda = par['lda']  # length/taup ratio 0.5 * np.sqrt(p)
    sigmas = par['sigmas']  # shear force, typically 10 .. 45 kNm/m²

    # airgap diameter
    Da = (4*p * tnom/(np.pi**2 * lda * sigmas)) ** (1./3)

    # pole width
    taup = np.pi * Da/(2*p)
    lfe = taup*lda

    if 'udc' in par:
        u1nom = 0.9*par['udc']/np.sqrt(2)/np.sqrt(3)  # phase voltage
    else:
        u1nom = par['u1']  # phase voltage

    eta = par['eta']  # efficiency
    cos_phi = par['cos_phi']

    # winding and poles
    m = par['m']  # num phases
    if 'Q1' not in par:
        Q1 = _stator_slots(par, slots)
    else:
        Q1 = par['Q1']
    coil_span = par.get('coil_span', 0)
    if np.gcd(Q1, 2*p) == coil_span:
        coil_span = 0
    layers = 1 if coil_span == 0 else 2
    req_poles = np.gcd(Q1, 2*p)
    if req_poles != 2*p:
        layers = 2

    check_symmetry_conditions(Q1, p, layers, m)

    Ui_U = par['ui_u']

    # design parameters
    # AJ = papyr['AJ'] # 100*1e9 # thermal load, typically 100 .. 300 A²/mm³
    J = par['J']   # current density 3 .. 6 A/mm²
    Ba = par['Ba']     # airgap flux density, typically 0.6 .. 1.5 T
    kq = par['kq']      # winding fill factor, typically 0.35 .. 0.75
    # flux density in teeth, typically 1.7 .. 2.2 T
    Bth = par['Bth']
    By = par['By']     # flux density in yoke, typically 1.2 .. 1.5 T

    # winding factor
    Nt = Q1/np.gcd(Q1, p)
    qt = Nt / (2*m)  # if Nt % 2 == 0 else Nt / m
    kwz = np.sin(np.pi/(2*m))/(qt*np.sin(np.pi/(2*m*qt)))
    kws = np.sin(np.pi*p*coil_span/Q1) if coil_span != 0 else 1.0
    kw = kwz * kws

    # flux density amplitude of base harmonic
    if 'brem' in par:
        mag_width = par['mag_width']
        Bd1 = 4.0/np.pi*Ba*np.sin(np.pi/2.0*mag_width)
    else:
        Bd1 = Ba
   # ---------------------------------------------
    f1nom = speednom*p

    # stator winding
    psi1 = 2.0/np.pi*taup*lfe*Bd1

    Ui = Ui_U * u1nom
    N = round(np.sqrt(2)*Ui/(2*np.pi*f1nom*kw*psi1), 1)

    # current loading
    # A = np.sqrt(2)*sigmas/kw/Ba
    I1 = pnom/(m*eta*cos_phi*u1nom)
    A = 2*m*N*I1/np.pi/Da

    # slot area
    # J = AJ/A
    hs1 = 1e-3  # slot opening height
    taus = np.pi/Q1
    ans = taus*Da*A/(kq*J)
    bds = taus*(Da+2*hs1)*Bd1/Bth
    bns = taus*(Da+2*hs1) - bds

    hns = (-bns + np.sqrt(bns**2 + 4*ans*np.tan(taus)))/2/np.tan(taus)
    hys = psi1/2/lfe*By

    # airgap and yoke diameter
    airgap = par['airgap']
    if par['external_rotor']:
        Da1 = Da - 2*airgap
        Dy1 = Da1 - 2*(hys + hns + hs1)
        Da2 = Da
    else:
        Da1 = Da
        Dy1 = Da1 + 2*(hys + hns + hs1)
        Da2 = Da1-2*airgap
    r = dict(
        lfe=round(lfe, 3),
        Dy1=round(Dy1, 3),
        Da1=round(Da1, 4),
        Da2=round(Da2, 4),
        ans=round(ans, 6),
        hns=round(hns, 4),
        bns=round(bns, 4),
        A=round(A, 3),
        AJ=round(J*A, 0),
        ess=round(1e-3*pnom/(60*speednom)/(Da1**2*lfe), 4),
        trv=round(1e-3*pnom/(2*np.pi*speednom)/(np.pi*Da1**2/4*lfe), 1),
        w1=int(N),
        kw=round(kw, 4),
        q=qt,
        i1=round(I1, 3),  # np.pi*Da1*A/2/m/N
        psi1=round(psi1, 5))

    slotwidth = np.floor(0.3*np.pi*Da1/Q1*2000+0.5)/2000.0
    tw = bds  # np.pi*(Da1+hns)/Q1 - bns
    toothwidth = np.floor(tw*2000+0.5)/2000.0
    bns1 = np.floor((np.pi*Da1/Q1 - tw)*2000.0+0.5)/2000.0
    bns2 = np.floor(((np.pi*Da1+2*hns)/Q1 - tw)*2000+0.5)/2000.0

    r['pfe'] = round(iron_losses(Dy1, lfe, (Dy1/2-Da1/2-hns),
                                 toothwidth, hns, Q1,
                                 Bd1, f1nom), 1)

    if coil_span == 0:
        tp = Q1/2./p
        q = tp/m
        coil_span = Q1/p/2 if q >= 1 else 1

    middle_line = 1 if coil_span == 1 else 2

    # stator 3
    r['stator'] = dict(
        u1nom=round(u1nom, 1), f1nom=round(f1nom, 1),
        num_slots=Q1,
        nodedist=1,
        # num_slots_gen = req_poles*Q1/2/p,
        statorRotor3=dict(
            slot_width=slotwidth,
            tooth_width=toothwidth,
            slot_height=hs1+r['hns'],
            slot_top_sh=1.0,
            slot_h1=hs1,
            slot_h2=hs1,
            slot_r1=1.5e-3,
            slot_r2=1.5e-3,  # bns2/2,
            wedge_width1=0,  # bns1,
            wedge_width2=0,
            middle_line=0 if layers < 2 else middle_line))
    # coils per phase and pole
    # q = Q/2/p/m
    # num wires per coil side
    # n = a*N / 2 / p / q
    num_wires = round(2*m*N/layers/Q1)
    dwire = 2*np.sqrt(ans*kq/layers/np.pi/num_wires)
    relculen = 1.4
    r1 = wdg_resistance(N, relculen*lfe, dwire)
    r['pcu'] = round(m*r1*I1**2, 1)

    r['winding'] = dict(
        wire_diam=round(dwire, 5),
        num_phases=m,
        cufilfact=kq,
        culength=1.4,
        num_layers=layers,
        resistance=round(r1, 4),
        coil_span=int(coil_span),
        num_wires=int(num_wires))

    return r


def _get_magnet_height(I1, N, kw, par):
    airgap = par['airgap']
    Ba = par['Ba']
    p = par['p']
    m = par['m']

    Hc = par['Hc']
    Hc = Hc*1e3  # unit kA/m
    # Safety Factor for demagnetization
    demag = par['demag']
    I1max = 4*I1
    THETA1 = m/np.pi*np.sqrt(2)*I1max*N*kw/p
    # minimal magnet height with safety factor
    hM = THETA1/Hc*demag
    Br = par['brem']
    if hM/(hM+airgap)*Br < Ba:
        raise ValidationError(
            f'Flux density in airgap {Ba:4.2f} T is too high, must be smaller than {hM/(hM+airgap)*Br:4.2f} T')
    return hM


def _get_magnet_yoke_diameter(psi1, lfe, Da2, hM, By, external_rotor):
    hys = psi1/(2*By*lfe)
    if external_rotor:
        return (Da2 + 1.2*2*hM + 2*hys)

    return (Da2 - 1.2*2*hM - 2*hys)


def get_surface_magnet_dimensions(I1, N, kw, psi1, lfe, Da2, par):
    hM = _get_magnet_height(I1, N, kw, par)
    Dy2 = _get_magnet_yoke_diameter(
        psi1, lfe, Da2, hM, par['By'], par['external_rotor'])
    return dict(
        Da2=np.floor(Da2*1e4+0.5)/1e4,
        Dy2=np.floor(Dy2*1e4+0.5)/1e4,
        hM=round(hM, 5),
        nodedist=1.0,
        magnetSector=dict(
            magn_height=np.floor(hM*2000+0.5)/2000.0,
            magn_width_pct=par['mag_width'],
            condshaft_r=round(Dy2/2, 4),
            magn_num=1,
            magn_rfe=0.0,
            magn_len=1.0,
            magn_shape=0.0,
            bridge_height=0.0,
            bridge_width=0.0,
            magn_ori=2,
            magn_type=1))


def get_interior_magnet_dimensions(I1, N, kw, psi1, lfe, Da2, par):
    r = dict()
    hM = round(_get_magnet_height(I1, N, kw, par), 5)
    Dy2 = _get_magnet_yoke_diameter(
        psi1, lfe, Da2, hM, par['By'], par['external_rotor'])

    kr = 0.9
    # pole pitch angle
    alphap = np.pi/par['p']
    hfe = par['hfe']
    tap = np.tan(alphap/2)
    sip = np.sin(alphap/2)
    a = (1+1/tap**2)
    b = 4/tap*(hfe/sip+hM)
    c = -Da2**2 + 4*Da2*hfe - 4*hfe**2 + 4*(hfe/sip+hM)**2
    wM = (-b+np.sqrt(b**2 - 4*a*c))/2/a
    hp = Da2/2 - (wM/tap/2 + hfe/sip + hM)

    # Magnet iron 4 BFE
    mag_dm_min = Da2/2 - np.sqrt((Da2**2 - wM**2)/4)
    mag_dm_max = (Da2-Dy2)/2 - hM
    magn_di_ra = mag_dm_min * kr + mag_dm_max * (1 - kr)
    mag_corner_center_distance = np.sqrt(
        (Da2/2-magn_di_ra-hM)**2 + (wM/2)**2)
    iron_bfe = (mag_corner_center_distance *
                np.sin(alphap/2 -
                       np.arcsin(wM/2/mag_corner_center_distance)))
    return dict(
        Da2=np.floor(Da2*1e4+0.5)/1e4,
        Dy2=np.floor(Dy2*1e4+0.5)/1e4,
        hM=hM,
        nodedist=1.0,
        magnetIron=dict(
            magn_height=np.floor(hM*2000+0.5)/2000.0,
            magn_width=round(wM, 4),
            air_triangle=1,
            magn_rem=par['brem'],
            iron_height=round(hfe, 4),
            gap_ma_iron=0,
            bridge_height=0.0,
            bridge_width=0.0,
            magn_ori=2,
            condshaft_r=round(Dy2/2, 4),
            iron_shape=Da2/2))


def get_im_rotor_dimensions(A, Da2, psi1, lfe, par, rtype='rotorKs2'):
    r = dict()
    r['Da2'] = Da2
    if 'Q2' not in par:
        r['num_slots'] = _rotor_slots(par['Q1'], par['p'])[0]
    else:
        r['num_slots'] = par['Q2']
    # slot area
    # J = AJ/A
    J = par['J']
    hs1 = 1e-3
    alfar = np.pi/r['num_slots']
    kq = par['kqr']
    Ar = alfar*Da2*A/J/kq
    wt = alfar*(Da2-2*hs1)*par['Ba']/par['Bth']
    wr = alfar*(Da2-2*hs1) - wt
    #hr = (wr + np.sqrt(wr**2 - 4 * Ar*np.tan(alphar)))(2*np.tan(alphar))
    hr = (wr + np.sqrt(wr**2 +
                       4*Ar*(1-2*np.tan(alfar))))/2/(1-2*np.tan(alfar))
    taup = np.pi * Da2/(2*par['p'])
    hyr = psi1/2/lfe*par['By']
    r['Dy2'] = round(Da2 - 2*hr - 2*hyr, 4)
    logger.info("Dy2 %f Da2 %f hys %f hr %f",
                r['Dy2']*1e3, Da2*1e3, hyr*1e3, hr*1e3)
    slotwidth = 1e-3
    Q2 = r['num_slots']
    r1 = wr/2-slotwidth
    r2 = (Da2/2-hr-hs1)*np.tan(alfar)
    if rtype == 'statorRotor3':
        r['statorRotor3'] = dict(
            slot_width=slotwidth,
            tooth_width=round(wt, 4),
            slot_height=round(hr+r2, 4),
            slot_top_sh=1.0,
            slot_h1=round(hs1, 4),
            slot_h2=round(hs1+r1, 4),
            slot_r1=round(r1, 4),
            slot_r2=round(r2),
            wedge_width1=0,
            wedge_width2=0,
            middle_line=0)
    elif rtype == 'rotorAsyn':
        r['rotorAsyn'] = dict(
            slot_bs2=0.1e-3,
            slot_hs2=0.5e-3,
            slot_b32=0.0,
            slot_h32=0.0,
            slot_b42=0.0,
            slot_h42=0.0,
            slot_b52=round(wr, 4),
            slot_b62=3e-3,
            slot_h52=2.5e-3,
            slot_h62=round(hr, 4),
            slot_h72=2e-3)
    else:
        r['rotorKs2'] = dict(
            slot_angle=round(2*alfar*180/np.pi, 2),
            slot_height=round(hr+r1+r2, 4),
            slot_topwidth=round(wr, 4),
            slot_width=slotwidth,
            slot_h1=hs1,
            slot_h2=0,
            slot_r1=1e-3,  # r1,
            slot_r2=1e-3,  # r2,
            middle_line=0)

    return r


def get_sm_rotor_dimensions(A, psi1, lfe, Da2, par):
    r = dict()
    r['Da2'] = Da2
    r['num_wires'] = 1

    alphap = np.pi/par['p']
    wp = 0.9*Da2*np.sin(alphap/2)
    hp = Da2/2*(1-np.cos(alphap/2))
    wc = alphap/2*Da2*par['Ba']/par['Bthr']
    hfmax = (wp-wc)/2/np.tan(alphap/2)
    mue0 = 4*np.pi*1e-7

    wf = wp - wc
    anr = par['airgap']*par['Ba']/mue0/par['J']/par['kqr']

    a = np.tan(alphap/2)/2
    b = (wp/2 - wc + np.tan(alphap/2)*(Da2/2 - hp))/2
    hc = min(max(4*anr/(wc+wp), 0.75*hfmax), hfmax)

    hyr = psi1/2/lfe*par['Byr']
    pr = 0.95*Da2/2

    r['Dy2'] = round(Da2 - 2*hp - 2*hyr - 2*hc, 4)
    if r['Dy2'] < 6e-3:
        hyr = hyr + r['Dy2']/2
        r['Dy2'] = 6e-3  # minimal size

    r['rot_hsm'] = dict(
        gap_pol_shaft=0.0,
        core_height=round(hc, 4),
        pole_height=round(hp, 4),
        pole_rad=round(pr, 4),
        core_width2=round(wc, 4),
        core_width1=round(wc, 4),
        pole_width_r=round(wp, 4),
        pole_width=round(wp, 4),
        slot_width=0,
        slot_height=0,
        damper_diam=0,
        damper_div=0
    )
    return r


def _set_defaults(par, defaults):
    for k in defaults:
        if k not in par:
            par[k] = defaults[k]


def _set_pm_defaults(par):
    _set_defaults(par, PM_DEFAULTS)


def _set_im_defaults(par):
    _set_defaults(par, IM_DEFAULTS)


def _set_sm_defaults(par):
    _set_defaults(par, SM_DEFAULTS)


def _set_genpars(r, poles):
    r.pop('Da2')
    for k in ('magnet', 'rotor'):
        if k in r:
            break
    r['outer_diam'] = r.pop('Dy1')
    r['inner_diam'] = r[k].pop('Dy2')
    r['airgap'] = (r['Da1'] - r[k].pop('Da2'))/2
    r['bore_diam'] = r.pop('Da1')
    r['poles'] = poles
    for k in 'u1nom', 'f1nom':
        r[k] = r['stator'].pop(k)


def spm(pnom: float, speed: float, p: int, **kwargs) -> dict:
    """returns dimension of a SPM machine

    Args:
    pnom: power at rated speed (W)
    speed: rotation speed (1/s)
    p: number of pole pairs

    udc: (optional) DC link voltage (V)
    u1: (optional) phase voltage (Vrms)
    Q1: (optional) total number of stator slots
    brem: (optional) remanence of magnet (T)
    """
    par = dict(
        pnom=pnom, speed=speed, p=p)
    par.update(kwargs)

    _set_pm_defaults(par)
    # stator and magnet parameters of surface mounted magnet machine
    r = get_stator_dimensions(par)

    # magnet parameters
    r['magnet'] = get_surface_magnet_dimensions(
        r['i1'], r['w1'], r['kw'], r['psi1'], r['lfe'], r['Da2'],
        par)

    r['magnet'].pop('hM')
    _set_genpars(r, 2*par['p'])
    r['name'] = f"SPM-{r['poles']}"

    return r


def ipm(pnom: float, speed: float, p: int, **kwargs) -> dict:
    """returns dimension of a IPM machine

    Args:
    pnom: power at rated speed (W)
    speed: rotation speed (1/s)
    p: number of pole pairs

    udc: (optional) DC link voltage (V)
    u1: (optional)  phase voltage (Vrms)
    Q1: (optional) total number of stator slots
    brem: (optional) remanence of magnet (T)

    """
    par = dict(
        pnom=pnom, speed=speed, p=p)
    par.update(kwargs)
    _set_pm_defaults(par)

    # stator and magnet parameters of interior mounted magnet machine
    r = get_stator_dimensions(par)

    # magnet parameters
    r['magnet'] = get_interior_magnet_dimensions(
        r['i1'], r['w1'], r['kw'], r['psi1'], r['lfe'], r['Da2'],
        par)

    r['magnet'].pop('hM')
    _set_genpars(r, 2*par['p'])
    r['name'] = f"IPM-{r['poles']}"

    return r


def im(pnom: float, speed: float, p: int, **kwargs) -> dict:
    """returns dimension of a IM machine

    Args:
    pnom: power at rated speed (W)
    speed: rotation speed (1/s)
    p: number of pole pairs

    udc: (optional) DC link voltage (V)
    u1: (optional) phase voltage (Vrms)
    Q1: (optional) total number of stator slots
    Q2: (optional) total number of rotor slots
    """
    par = dict(
        pnom=pnom, speed=speed, p=p)
    par.update(kwargs)
    _set_im_defaults(par)

    # stator and rotor parameters of induction machine
    try:
        slots = _im_stator_slots(p)
    except KeyError:
        slots = []
    r = get_stator_dimensions(par, slots=slots)
    # rotor parameters
    r['rotor'] = get_im_rotor_dimensions(
        par['cos_phi']*r['A'], r['Da2'], r['psi1'], r['lfe'], par)
    _set_genpars(r, 2*par['p'])
    r['name'] = f"IM-{r['poles']}"
    return r


def eesm(pnom: float, speed: float, p: int, **kwargs) -> dict:
    """returns dimension of a EESM machine

    Args:
    pnom: power at rated speed (W)
    speed: rotation speed (1/s)
    p: number of pole pairs

    udc: (optional) DC link voltage (V)
    u1: (optional) phase voltage (Vrms)
    Q1: (optional) total number of stator slots

    """
    par = dict(
        pnom=pnom, speed=speed, p=p)
    par.update(kwargs)
    _set_sm_defaults(par)

    # stator and rotor parameters of ee synchronous machine
    r = get_stator_dimensions(par)

    # rotor parameters
    r['rotor'] = get_sm_rotor_dimensions(r['A'], r['psi1'],
                                         r['lfe'], r['Da2'], par)

    _set_genpars(r, 2*par['p'])
    r['name'] = f"EESM-{r['poles']}"
    return r


if __name__ == "__main__":

    pnom = 10e3
    speed = 4400/60
    p = 4
    udc = 600

    r = spm(pnom, speed, p, udc=udc)

    # print results
    print("Number of slots          {:10d}".format(r['stator']['num_slots']))
    print("Iron length            [mm] {:10.2f}".format(r['lfe']*1e3))
    print("Stator outer diameter  [mm] {:10.2f}".format(r['outer_diam']*1e3))
    print("Stator bore diameter   [mm] {:10.2f}".format(r['bore_diam']*1e3))
    print("Slot height            [mm] {:10.2f}".format(
        r['stator']['statorRotor3']['slot_height']*1e3))
#    print("     width             [mm] {:10.2f}".format(
#        r['stator']['statorRotor3']['slot_']*1e3))
    print("Current loading      [kA/m] {:10.2f}".format(r['A']*1e-3))
    print("Thermal load       [A²/mm³] {:10.2f}".format(r['AJ']*1e-9))
    print("Esson           [kVAmin/m³] {:10.2f}".format(r['ess']))
#    print("Winding                    {:7}".format(r['winding']))
    print("Windingfactor                {:10.3f}".format(r['kw']))
    print("Slots/Pole Phase            {:9.1f}".format(r['q']))
    print("Current                 [A] {:10.2f}".format(r['i1']))
    print("Iron Losses             [W] {:10.2f}".format(r['pfe']))
    print("Cur Losses              [W] {:10.2f}".format(r['pcu']))
#    json.dump(r, sys.stdout)



# -*- coding: utf-8 -*-
"""
    femagtools.convert
    ~~~~~~~~~~~~~~~~~~

    Convert FEMAG files to various unstructured grid file formats
"""
import logging
import io
import numpy as np
from femagtools import isa7
from femagtools import nc
from collections import defaultdict, namedtuple

logger = logging.getLogger('femagtools.convert')


output_filetypes = [
    "msh", "geo", "vtu"]


def _from_isa(isa, filename, target_format,
              extrude=0, layers=0, recombine=False):

    try:
        airgap_outer_vertices = [v for e in isa.airgap_outer_elements
                                 for v in e.vertices]
    except AttributeError:
        logger.warning("airgap radius is not set in source file")

    airgap_lines = []
    for e in isa.airgap_center_elements:
        ev = e.vertices
        for i, v1 in enumerate(ev):
            v2 = ev[i-1]
            if v1 in airgap_outer_vertices and \
               v2 in airgap_outer_vertices:
                airgap_lines.append((v1, v2))

    nodechain_links = defaultdict(list)
    for n in isa.nodechains:
        nodechain_links[n.node1].extend(n.nodes)
        nodechain_links[n.node2].extend(n.nodes)
        if n.nodemid is not None:
            nodechain_links[n.nodemid].extend(n.nodes)

    physical_lines = ["v_potential_0",
                      "v_potential_const",
                      "periodic_+",
                      "periodic_-",
                      "infinite_boundary",
                      "no_condition",
                      "Airgap"]

    physical_surfaces = sorted(set([sr.name
                                    for sr in isa.subregions]
                                   + ["Winding_{}_{}".format(w.key, pol)
                                      for w in isa.windings
                                      for pol in ("+", "-")]
                                   + ["Air_Inner",
                                      "Air_Outer",
                                      "Airgap_Inner",
                                      "Airgap_Outer",
                                      "PM1", "PM2",
                                      "PM3", "PM4"]))

    def physical_line(n1, n2):
        if (n1, n2) in airgap_lines or (n2, n1) in airgap_lines:
            return 7  # airgap
        if n1.bndcnd == n2.bndcnd:
            return boundary_condition(n1)
        if boundary_condition(n1) == 1:
            return boundary_condition(n2)
        return boundary_condition(n1)

    def boundary_condition(node):
        if node.bndcnd == 0:
            return 6  # no condition
        if node.bndcnd == 1:
            return 1  # vpot 0
        if node.bndcnd == 2:
            return 2  # vpot const
        if node.bndcnd == 3 or node.bndcnd == 6:
            return 4  # periodic -
        if node.bndcnd == 4 or node.bndcnd == 5:
            return 3  # periodic +
        if node.bndcnd == 8 or node.bndcnd == 9:
            return 1  # vpot 0

    def physical_surface(e):

        def surface_id(name):
            return physical_surfaces.index(name) + len(physical_lines) + 1

        if any(e.mag):
            if e.mag[0] > 0:
                if e.mag[1] > 0:
                    return surface_id("PM1")
                return surface_id("PM2")
            if e.mag[1] > 0:
                return surface_id("PM3")
            return surface_id("PM4")

        if e in isa.airgap_inner_elements or e in isa.airgap_center_elements:
            return surface_id("Airgap_Inner")

        if e in isa.airgap_outer_elements:
            return surface_id("Airgap_Outer")

        sr_key = isa.superelements[e.se_key].sr_key
        if sr_key == -1:
            if e.vertices[0].outside:
                return surface_id("Air_Outer")
            return surface_id("Air_Inner")

        sr = isa.subregions[sr_key]
        if sr.wb_key != -1:
            wb = isa.subregions[sr.wb_key]
            if sr.curdir > 0:
                return surface_id("Winding_{}_-".format(wb.key))
            return surface_id("Winding_{}_+".format(wb.key))

        return surface_id(sr.name)

    def line_on_boundary(n1, n2):
        if n1.on_boundary() and n2.on_boundary():
            if n1 in nodechain_links.keys():
                return n2 in nodechain_links[n1]
            else:
                return False
        else:
            return (n1, n2) in airgap_lines or (n2, n1) in airgap_lines

    points = [[n.x, n.y, 0] for n in isa.nodes]
    vpot = [n.vpot[0] for n in isa.nodes]

    lines = []
    line_ids = []
    triangles = []
    quads = []
    physical_ids = dict(triangle=[], quad=[])
    geometrical_ids = dict(triangle=[], quad=[])
    b = dict(triangle=[], quad=[])
    h = dict(triangle=[], quad=[])
    perm = dict(triangle=[], quad=[])
    iron_losses = dict(triangle=[], quad=[])
    mag_losses = dict(triangle=[], quad=[])
    wdg_losses = dict(triangle=[], quad=[])
    for e in isa.elements:
        ev = e.vertices

        for i, v in enumerate(ev):
            v1, v2 = v, ev[i-1]
            if line_on_boundary(v1, v2):
                lines.append([v1.key - 1, v2.key - 1])
                line_ids.append(physical_line(v1, v2))

        if len(ev) == 3:
            triangles.append([n.key - 1 for n in ev])
            cell_type = "triangle"

        elif len(ev) == 4:
            quads.append([n.key - 1 for n in ev])
            cell_type = "quad"

        try:
            magtemp = isa.MAGN_TEMPERATURE
        except AttributeError:
            magtemp = 20

        physical_ids[cell_type].append(physical_surface(e))
        geometrical_ids[cell_type].append(e.se_key)
        b[cell_type].append(e.flux_density())
        h[cell_type].append(e.demagnetization(magtemp))
        perm[cell_type].append(e.permeability())
        iron_losses[cell_type].append(e.iron_loss_density())
        mag_losses[cell_type].append(e.mag_loss_density())
        wdg_losses[cell_type].append(e.wdg_loss_density())

    logger.info("%s: Lines %d, Triangles %d, Quads %d",
                filename, len(lines), len(triangles), len(quads))

    if target_format == "msh":
        import meshio

        points = np.array(points)
        point_data = {"potential": np.array(vpot)}

        cells = []
        cell_data = defaultdict(list)

        if lines:
            cells.append(("line", np.array(lines)))
            cell_data["gmsh:geometrical"].append(np.array(line_ids))
            cell_data["gmsh:physical"].append(np.array(line_ids))
            cell_data["b"].append(np.zeros((len(lines), 3)))
            cell_data["h"].append(np.zeros(len(lines)))
            cell_data["Rel. Permeability"].append(np.zeros(len(lines)))
            cell_data["Iron Loss Dens."].append(np.zeros(len(lines)))
            cell_data["Mag. Loss Dens."].append(np.zeros(len(lines)))
            cell_data["Wdg. Loss Dens."].append(np.zeros(len(lines)))

        if triangles:
            cells.append(("triangle", np.array(triangles)))
            cell_data["gmsh:geometrical"].append(np.array(geometrical_ids["triangle"]))
            cell_data["gmsh:physical"].append(np.array(physical_ids["triangle"]))
            cell_data["b"].append(np.array([i + (0,) for i in b["triangle"]]))
            cell_data["h"].append(np.array(h["triangle"]))
            cell_data["Rel. Permeability"].append(np.array(perm["triangle"]))
            cell_data["Iron Loss Dens."].append(np.array(iron_losses["triangle"]))
            cell_data["Mag. Loss Dens."].append(np.array(mag_losses["triangle"]))
            cell_data["Wdg. Loss Dens."].append(np.array(wdg_losses["triangle"]))

        if quads:
            cells.append(("quad", np.array(quads)))
            cell_data["gmsh:geometrical"].append(np.array(geometrical_ids['quad']))
            cell_data["gmsh:physical"].append(np.array(physical_ids['quad']))
            cell_data["b"].append(np.array([i + (0,) for i in b['quad']]))
            cell_data["h"].append(np.array(h['quad']))
            cell_data["Rel. Permeability"].append(np.array(perm['quad']))
            cell_data["Iron Loss Dens."].append(np.array(iron_losses['quad']))
            cell_data["Mag. Loss Dens."].append(np.array(mag_losses['quad']))
            cell_data["Wdg. Loss Dens."].append(np.array(wdg_losses['quad']))

        field_data = {}
        for l in physical_lines:
            field_data[l] = np.array([physical_lines.index(l) + 1, 1])
        for s in physical_surfaces:
            field_data[s] = np.array([physical_surfaces.index(s) + 1
                                      + len(physical_lines), 2])
        meshio.write_points_cells(filename,
                                  points,
                                  cells,
                                  point_data,
                                  cell_data,
                                  field_data,
                                  file_format="gmsh22",
                                  binary=False)

    if target_format == "geo":
        import meshio
        geo = []
        nc_nodes = set([n for c in isa.nodechains for n in c.nodes])

        for n in isa.nodes:
            if n in nc_nodes:
                geo.append("Point({}) = {{{}, {}, {}}};".format(
                    n.key, n.x, n.y, 0))

        for c in isa.nodechains:
            geo.append("Line({}) = {{{}}};".format(
                c.key, ", ".join([str(n.key) for n in c.nodes])))
        used = set()
        for c in isa.nodechains:
            n1, n2 = c.nodes[0], c.nodes[1]
            if line_on_boundary(n1, n2):
                id_ = physical_line(n1, n2)
                name = physical_lines[id_ - 1]
                if extrude:
                    geo.append(
                        "extrusion[] = Extrude {{0, 0, {}}} {{ Line{{{}}}; {}{}}};".format(
                            extrude,
                            c.key,
                            "Layers{{{}}}; ".format(layers) if layers else "",
                            "Recombine; " if recombine else ""))

                    geo.append("Physical Surface('{}', {}) {} extrusion[1];".format(
                        name,
                        id_,
                        "+=" if name in used else "="))
                else:
                    geo.append("Physical Line('{}', {}) {} {{{}}};".format(
                        name,
                        id_,
                        "+=" if name in used else "=",
                        c.key))
                used.add(name)

        for se in isa.superelements:
            geo.append("Line Loop({}) = {{{}}};".format(
                se.key,
                ", ".join([str(c.key) for c in se.nodechains])))
            geo.append("Plane Surface({0}) = {{{0}}};".format(se.key))
        used = set()
        for se in isa.superelements:
            id_ = physical_surface(se.elements[0]) - len(physical_lines)
            name = physical_surfaces[id_ - 1]
            if extrude:
                geo.append(
                    "extrusion[] = Extrude {{0, 0, {}}} {{ Surface{{{}}}; {}{}}};".format(
                        extrude,
                        se.key,
                        "Layers{{{}}}; ".format(layers) if layers else "",
                        "Recombine; " if recombine else ""))

                geo.append("Physical Surface('base', {}) {} {{{}}};".format(
                    len(physical_lines) + 1,
                    "=" if se.key == 1 else "+=",
                    se.key))

                geo.append("Physical Surface('top', {}) {} extrusion[0];".format(
                    len(physical_lines) + 2,
                    "=" if se.key == 1 else "+="))

                geo.append("Physical Volume('{}', {}) {} extrusion[1];".format(
                    name,
                    id_,
                    "+=" if name in used else "="))
            else:
                geo.append("Physical Surface('{}', {}) {} {{{}}};".format(
                    name,
                    id_,
                    "+=" if name in used else "=",
                    se.key))
            used.add(name)

        with open(filename, "w") as f:
            f.write("\n".join(geo))

    if target_format == "vtu":
        import meshio
        assert len(points) == len(vpot)
        assert len(lines) == len(line_ids)
        assert len(triangles) == len(physical_ids['triangle']) == len(geometrical_ids['triangle']) == len(b['triangle']) == len(h['triangle']) == len(perm['triangle'])
        assert len(quads) == len(physical_ids['quad']) == len(geometrical_ids['quad']) == len(b['quad']) == len(h['quad']) == len(perm['quad'])

        points = np.array(points)
        point_data = {"potential": np.array(vpot)}

        cells = []
        cell_data = defaultdict(list)

        if lines:
            cells.append(("line", np.array(lines)))
            cell_data["GeometryIds"].append(np.array(line_ids))
            cell_data["PhysicalIds"].append(np.array(line_ids))
            cell_data["b"].append(np.zeros((len(lines), 3)))
            cell_data["Demagnetization"].append(np.zeros(len(lines)))
            cell_data["Rel. Permeability"].append(np.zeros(len(lines)))
            cell_data["Iron Loss Dens."].append(np.zeros(len(lines)))
            cell_data["Mag. Loss Dens."].append(np.zeros(len(lines)))
            cell_data["Wdg. Loss Dens."].append(np.zeros(len(lines)))

        if triangles:
            cells.append(("triangle", np.array(triangles)))
            cell_data["GeometryIds"].append(np.array(geometrical_ids['triangle']))
            cell_data["PhysicalIds"].append(np.array(physical_ids['triangle']))
            cell_data["b"].append(np.array([i + (0,) for i in b['triangle']]))
            cell_data["Demagnetization"].append(np.array(h['triangle']))
            cell_data["Rel. Permeability"].append(np.array(perm['triangle']))
            cell_data["Iron Loss Dens."].append(np.array(iron_losses['triangle']))
            cell_data["Mag. Loss Dens."].append(np.array(mag_losses['triangle']))
            cell_data["Wdg. Loss Dens."].append(np.array(wdg_losses['triangle']))

        if quads:
            cells.append(("quad", np.array(quads)))
            cell_data["GeometryIds"].append(np.array(geometrical_ids['quad']))
            cell_data["PhysicalIds"].append(np.array(physical_ids['quad']))
            cell_data["b"].append(np.array([i + (0,) for i in b['quad']]))
            cell_data["Demagnetization"].append(np.array(h['quad']))
            cell_data["Rel. Permeability"].append(np.array(perm['quad']))
            cell_data["Iron Loss Dens."].append(np.array(iron_losses['quad']))
            cell_data["Mag. Loss Dens."].append(np.array(mag_losses['quad']))
            cell_data["Wdg. Loss Dens."].append(np.array(wdg_losses['quad']))

        field_data = {}
        for l in physical_lines:
            field_data[l] = np.array([physical_lines.index(l) + 1, 1])
        for s in physical_surfaces:
            field_data[s] = np.array([physical_surfaces.index(s) + 1
                                      + len(physical_lines), 2])
        meshio.write_points_cells(filename,
                                  points,
                                  cells,
                                  point_data=point_data,
                                  cell_data=cell_data,
                                  field_data=field_data,
                                  file_format="vtu",
                                  binary=True)


def _from_jmag(designer):
    """ Designer.jplot: plotmesh.data
    (older version Designer.jcf: mesh.dat
    """
    from zipfile import ZipFile
    def get_recs():
        with ZipFile(designer) as jplt:
            with jplt.open('plotmesh.dat') as f:
                numsecs = 0
                seclen = 0
                nline = 0
                for line in io.TextIOWrapper(f, 'utf-8'):
                    rec = line.strip().split()
                    nline += 1
                    if seclen == 0 and (rec[-1]=='*Node' or rec[-1]=='*Element'):
                        seclen = int(rec[0])
                        nline = 0
                        if numsecs>0:
                            yield []
                    elif nline < seclen:
                        yield rec
                    else:
                        if numsecs > 1:
                            break
                        seclen = 0
                        numsecs += 1
                        yield rec

    Node = namedtuple("Node", ["id", "x", "y"])
    nodes = []
    Element = namedtuple("Element", ["nodes", "entity"])
    elements = []
    state = 'Node'
    for rec in get_recs():
        if rec == []:
            state = 'Element'
            continue
        if state == 'Node':
            nodes.append(Node(rec[1], rec[2], rec[3]))
        else:
            if rec[1] == "3":
                elements.append(Element([rec[-3], rec[-2], rec[-1], None], rec[2]))
            elif rec[1] == "4":
                elements.append(Element([rec[-4], rec[-3], rec[-2], rec[-1]], rec[2]))

    replace = {}
    for i, n1 in enumerate(nodes):
        for n2 in nodes[i+1:]:
            if n1.x == n2.x and n1.y == n2.y and n1.id != n2.id:
                if n1.id not in replace:
                    replace[n2.id] = n1.id

    nodes = [n for n in nodes if n not in replace]

    for e in elements:
        for i, n in enumerate(e.nodes):
            if n in replace:
                e.nodes[i] = replace[n]
    entities = set([e.entity for e in elements])
    logger.info("%s: nodes %d elements %d physical names %d",
                designer, len(nodes), len(elements), len(entities))
    physicalgroups = {}
    msh = ["$MeshFormat\n2.2 0 8\n$EndMeshFormat"]
    msh.append("$PhysicalNames")
    msh.append(f"{len(entities)}")
    for i, g in enumerate(entities):
        msh.append(f'1 {i+1} "{g}"')
        physicalgroups[g] = i+1
    msh.append("$EndPhysicalNames")
    msh.append("$Nodes")
    msh.append(str(len(nodes)))
    for i, n in enumerate(nodes):
        msh.append("{} {} {} 0.0".format(n.id, n.x, n.y))
    msh.append("$EndNodes")

    msh.append("$Elements")
    msh.append(str(len(elements)))
    for i, e in enumerate(elements):
        if e.nodes[3]:
            msh.append(
                "{} 3 2 {} {} {} {} {} {}".format(
                    i + 1, physicalgroups[e.entity], e.entity,
                    e.nodes[0], e.nodes[1], e.nodes[2], e.nodes[3]))
        else:
            msh.append(
                "{} 2 2 {} {} {} {} {}".format(
                    i + 1, physicalgroups[e.entity], e.entity,
                    e.nodes[0], e.nodes[1], e.nodes[2]))
    msh.append("$EndElements")
    return msh


def _from_nastran(source, filename):
    import meshio
    def _nastran_real_to_float(s):
        s = s.strip()
        if "E" not in s:
            s = s[0] + s[1:].replace("+", "e+").replace("-", "e-")
        return float(s)

    points = []
    point_index = 0
    point_indexes = {}
    cells = {"triangle": [], "quad": []}
    shell_ids = {"triangle": [], "quad": []}
    shell_material_ids = {}

    with open(source) as file:
        while True:
            line = file.readline()
            if not line or line.strip().startswith("BEGIN BULK"):
                break

        while True:
            line = file.readline()
            if not line or line.startswith("ENDDATA"):
                break

            if line.strip().startswith("$"):
                continue

            fields = [line[i:i+8] for i in range(0, len(line), 8)]

            if fields[0].startswith("GRID"):
                point_indexes[int(fields[1])] = point_index
                point_index += 1
                points.append([_nastran_real_to_float(f) / 1000 for f in fields[3:6]])

            elif fields[0].startswith("CTRIA3"):
                cells["triangle"].append([point_indexes[int(f)] for f in fields[3:6]])
                shell_ids["triangle"].append(int(fields[2]))

            elif fields[0].startswith("CQUAD4"):
                cells["quad"].append([point_indexes[int(f)] for f in fields[3:7]])
                shell_ids["quad"].append(int(fields[2]))

            elif fields[0].startswith("PSHELL"):
                shell_material_ids[int(fields[1])] = int(fields[2])

    meshio_cells = []
    meshio_cell_data = defaultdict(list)

    if cells["triangle"]:
        meshio_cells.append(("triangle", np.array(cells["triangle"])))
        meshio_cell_data["gmsh:geometrical"].append(np.array(shell_ids["triangle"]))
        meshio_cell_data["gmsh:physical"].append(np.array(shell_ids["triangle"]))

    if cells["quad"]:
        meshio_cells.append(("quad", np.array(cells["quad"])))
        meshio_cell_data["gmsh:geometrical"].append(np.array(shell_ids["quad"]))
        meshio_cell_data["gmsh:physical"].append(np.array(shell_ids["quad"]))

    meshio.write_points_cells(
        filename,
        np.array(points),
        meshio_cells,
        cell_data=meshio_cell_data,
        field_data={f"{str(n)}": (n, 2) for n in sorted(list(set(shell_material_ids)))},
        file_format="gmsh22",
        binary=False)


def to_msh(source, filename, infile_type=None):
    """
    Convert a FEMAG, NASTRAN or JMAG Model Input File to msh format.

    Arguments:
        source: instance of femagtools.isa7.Isa7 or name of I7/ISA7/NAS/JPLOT file
        filename: name of converted file
        infile_type: format of source file
    """
    if isinstance(source, isa7.Isa7):
        _from_isa(source, filename, "msh")

    elif type(source) == str:
        if infile_type:
            file_ext = infile_type.lower()
        else:
            file_ext = source.split(".")[-1].lower()

        if file_ext in ["isa7", "i7", "nc"]:
            isa = nc.read(source) if file_ext == 'nc' else isa7.read(source)
            _from_isa(isa, filename, "msh")
        elif file_ext == "nas":
            _from_nastran(source, filename)
        elif file_ext == "jplot":
            msh = _from_jmag(source)
            with open(filename, 'w') as fp:
                fp.write('\n'.join(msh))
        else:
            raise ValueError(
                "cannot convert files of format {} to .msh".format(file_ext))
    else:
        raise ValueError("cannot convert {} to .msh".format(source))


def to_geo(source, filename, extrude=0, layers=0,
           recombine=False, infile_type=None):
    """
    Convert a femag model to geo format.

    Arguments:
        source: instance of isa7.Isa7 or name of an I7/ISA7 file
        filename: name of converted file
        extrude: extrude surfaces using a translation along the z-axis
        layers: number of layers to create when extruding
        recombine: when extruding, recombine triangles into quadrangles and
                 tetraedra into to prisms, hexahedra or pyramids
        infile_type: format of source file
    """
    if isinstance(source, isa7.Isa7):
        _from_isa(source, filename, "geo", extrude, layers, recombine)

    elif type(source) == str:
        if infile_type:
            file_ext = infile_type.lower()
        else:
            file_ext = source.split(".")[-1].lower()

        if file_ext in ["isa7", "i7", "nc"]:
            isa = nc.read(source) if file_ext == 'nc' else isa7.read(source)
            _from_isa(isa, filename, "geo", extrude, layers, recombine)
        else:
            raise ValueError(
                "cannot convert files of format {} to .geo".format(file_ext))
    else:
        raise ValueError("cannot convert {} to .geo".format(source))


def to_vtu(source, filename, infile_type=None):
    """
    Convert a femag model to vtu format.

    Arguments:
        source: instance of isa7.Isa7 or name of an I7/ISA7 or nc file
        filename: name of converted file
        infile_type: format of source file
    """
    if isinstance(source, isa7.Isa7):
        _from_isa(source, filename, "vtu")

    elif type(source) == str:
        if infile_type:
            file_ext = infile_type.lower()
        else:
            file_ext = source.split(".")[-1].lower()

        if file_ext in ["isa7", "i7", "nc"]:
            isa = nc.read(source) if file_ext == 'nc' else isa7.read(source)
            _from_isa(isa, filename, "vtu")
        else:
            raise ValueError(
                "cannot convert files of format {} to .vtu".format(file_ext))
    else:
        raise ValueError("cannot convert {} to .vtu".format(source))


def main(argv=None):
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')
    # Parse command line arguments.
    parser = _get_parser()
    args = parser.parse_args(argv)

    if not args.output_format:
        args.output_format = args.outfile.split('.')[-1]

    if args.output_format == 'msh':
        to_msh(args.infile, args.outfile)
    elif args.output_format == 'geo':
        to_geo(args.infile, args.outfile,
               extrude=args.extrude, layers=args.layers,
               recombine=args.recombine)
    elif args.output_format == 'vtu':
        to_vtu(args.infile, args.outfile)
    else:
        raise ValueError(
                "unsupported output format {}".format(args.output_format))
    return


def _get_parser():
    """Parse input options."""
    import argparse
    import sys
    from .__init__ import __version__

    parser = argparse.ArgumentParser(description=("Convert to mesh formats."))

    parser.add_argument("infile", type=str, help="mesh file to be read from")

    parser.add_argument(
        "--output-format",
        "-o",
        type=str,
        choices=output_filetypes,
        help="output file format",
        default=None,
    )

    parser.add_argument(
        "--extrude",
        "-x",
        type=float,
        help="extrusion length",
        default=None,
    )

    parser.add_argument(
        "--layers",
        "-l",
        type=int,
        help="number of layers",
        default=None,
    )

    parser.add_argument(
        "--recombine",
        "-r",
        action='store_true',
        help="recombine")

    parser.add_argument("outfile", type=str, help="mesh file to be written to")

    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version="%(prog)s {}, Python {}".format(__version__, sys.version),
        help="display version information",
    )

    return parser


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')
    main()



#!/usr/bin/env python
#
import os
import femagtools.jhb


def test_read_jhb():
    testPath = os.path.split(__file__)[0]
    if not testPath:
        testPath = '.'
    filename = "data/M270-50A_1000Hz_L.jhb"
    jhb = femagtools.jhb.Reader('{0}/{1}'.format(testPath, filename))
    assert jhb['name'] == 'M270-50A_1000Hz_L'
    assert len(jhb['curve']) == 1
    assert len(jhb['curve'][0]['hi']) == 13
    assert jhb['curve'][0]['bi'][-1] == 1.17


def test_read_jhb_aniso():
    testPath = os.path.split(__file__)[0]
    if not testPath:
        testPath = '.'
    filename = "data/aniso.jhb"
    jhb = femagtools.jhb.Reader('{0}/{1}'.format(testPath, filename))
    assert jhb['name'] == 'aniso'
    assert len(jhb['curve']) == 2
    assert len(jhb['curve'][0]['hi']) == 39
    assert jhb['curve'][0]['bi'][-1] == 4.485




# -*- coding: utf-8 -*-
"""
    femagtools.magnet
    ~~~~~~~~~~~~~~~~~

    Creating and managing magnet material



"""


class Magnet:
    def __init__(self, mlist=[]):
        """initialize this object from a list of magnet objects"""
        self.magnets = {}
        for m in mlist:
            if 'id' in m:
                self.magnets[str(m['id'])] = m
            elif 'name' in m:
                self.magnets[m['name']] = m

    def find(self, id):
        """find magnet by id or name"""
        try:
            return self.magnets[id]
        except ValueError:
            pass  # not found
        except KeyError:
            pass
        return self.find_by_name(id)

    def find_by_name(self, name):
        """find magnet by name"""
        try:
            for k in self.magnets.keys():
                if self.magnets[k]['name'] == name:
                    return self.magnets[k]
            # not found
        except KeyError:
            pass
        return {}

    def __str__(self):
        return str([self.magnets[m]
                    for m in self.magnets])



"""
Postprocessing FEMAG-ME results
"""
__author__ = 'dapu zhang'

import re 
import numpy as np 
import matplotlib.pyplot as plt
import logging

#  set logging
logger = logging.getLogger(__name__)

def get_eigenvectors(filelist, psfiles):
    """return eigenfrequency and eigenvectors (raw data and psfiles)"""
    eigenvecs = {}
    eigenfreq = []
    ps_data = []
    temp = [] 
    temp_arr = []
    read_fig = False

    if len(psfiles) > 0: read_fig = True
    if len(filelist) == 0: return []
    
    for k, kk in zip(filelist, range(len(filelist))): 
        # read ps file
        if read_fig: ps_data.append(plt.imread(psfiles[kk]))
        eigenvecs[str(kk)] = {}
        with open(k, 'r') as f: 
            data = f.read().split('\n')
        eigenval = float(re.split('\s+', data[0].strip())[2])
        eigenfreq.append(eigenval/2/np.pi)
        
        for i in range(len(data[1:-1])): 
            for j in data[2+i].strip().split(' '): 
                if j != '':
                    temp.append(float(j))

        temp_arr = np.array(temp).reshape(-1, 5)
        idx_collect = []
        # remove the index, at which the dof equals zero
        for c in range(temp_arr.shape[0]): 
            if np.abs(temp_arr[c, 1]) < 1e-15 and \
            np.abs(temp_arr[c, 2]) < 1e-15: 
                idx_collect.append(c)
        # prepare the data from the further calculation
        temp_arr = np.delete(temp_arr, np.array(idx_collect), axis=0)
        temp_arr = np.insert(temp_arr, 3, np.zeros((temp_arr.shape[0], )), axis=1)
        temp_arr = np.hstack((temp_arr, np.zeros((temp_arr.shape[0], 1))))
        eigenvecs[str(kk)].update({'eigenvecs': temp_arr[:, 1:4]})
        # return node position
        if kk == 0: eigenvecs.update({'node_pos': temp_arr[:, 4::]})

    return [ps_data, eigenfreq, eigenvecs] 